import argparse
import logging
from time import strftime

import torch.distributed as dist
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler
import yaml
from torch.utils.tensorboard import SummaryWriter

import ModelOpt.AutoYOLO.admm
import ModelOpt.AutoYOLO.test  # import test.py to get mAP after each epoch
from ModelOpt.AutoYOLO.admm import GradualWarmupScheduler
from ModelOpt.AutoYOLO.models import *
from ModelOpt.AutoYOLO.check_compression import test_sparsity
from ModelOpt.AutoYOLO.utils.datasets import *
from ModelOpt.AutoYOLO.utils.utils import *




def train(hyp):
    global mixed_precision
    try:  # Mixed precision training https://github.com/NVIDIA/apex
        from apex import amp
    except:
        print('Apex recommended for faster mixed precision training: https://github.com/NVIDIA/apex')
        mixed_precision = False  # not installed

    wdir = 'weights' + os.sep  # weights dir
    last = wdir + 'last.pt'
    best = wdir + 'best.pt'
    results_file = './results.txt'

    import pkg_resources
    PRUNE_CONFIG_PATH = pkg_resources.resource_filename('ModelOpt.AutoYOLO.train', 'prune_config/'+ opt.config_file + ".yaml")

    cfg = opt.cfg
    data = opt.data
    epochs = opt.epochs  # 500200 batches at bs 64, 117263 images = 273 epochs
    batch_size = opt.batch_size
    accumulate = max(round(64 / batch_size), 1)  # accumulate n times before optimizer update (bs 64)
    weights = opt.weights  # initial training weights
    imgsz_min, imgsz_max, imgsz_test = opt.img_size  # img sizes (min, max, test)

    # Image Sizes
    gs = 32  # (pixels) grid size
    assert math.fmod(imgsz_min, gs) == 0, '--img-size %g must be a %g-multiple' % (imgsz_min, gs)
    opt.multi_scale |= imgsz_min != imgsz_max  # multi if different (min, max)
    if opt.multi_scale:
        if imgsz_min == imgsz_max:
            imgsz_min //= 1.5
            imgsz_max //= 0.667
        grid_min, grid_max = imgsz_min // gs, imgsz_max // gs
        imgsz_min, imgsz_max = int(grid_min * gs), int(grid_max * gs)
    img_size = imgsz_max  # initialize with max size

    # Configure run
    init_seeds()
    data_dict = parse_data_cfg(data)
    train_path = data_dict['train']
    test_path = data_dict['valid']
    nc = 1 if opt.single_cls else int(data_dict['classes'])  # number of classes
    hyp['cls'] *= nc / 80  # update coco-tuned hyp['cls'] to current dataset

    # Remove previous results
    for f in glob.glob('*_batch*.jpg') + glob.glob(results_file):
        os.remove(f)

    # Initialize model
    model = Darknet(cfg).to(device)

    # Optimizer

    pg0, pg1, pg2 = [], [], []  # optimizer parameter groups
    for k, v in dict(model.named_parameters()).items():
        if '.bias' in k:
            pg2 += [v]  # biases
        elif 'Conv2d.weight' in k:
            pg1 += [v]  # apply weight_decay
        else:
            pg0 += [v]  # all else

    if opt.adam:
        # hyp['lr0'] *= 0.1  # reduce lr (i.e. SGD=5E-3, Adam=5E-4)
        optimizer = optim.Adam(pg0, lr=hyp['lr0'])
        # optimizer = AdaBound(pg0, lr=hyp['lr0'], final_lr=0.1)
    else:
        optimizer = optim.SGD(pg0, lr=hyp['lr0'], momentum=hyp['momentum'], nesterov=True)
    optimizer.add_param_group({'params': pg1, 'weight_decay': hyp['weight_decay']})  # add pg1 with weight_decay
    optimizer.add_param_group({'params': pg2})  # add pg2 (biases)
    print('Optimizer groups: %g .bias, %g Conv2d.weight, %g other' % (len(pg2), len(pg1), len(pg0)))
    del pg0, pg1, pg2

    start_epoch = 0
    best_fitness = 0.0
    # attempt_download(weights)

    
    if opt.freeze_layers:                                                                                                                                                            
        output_layer_indices = [idx - 1 for idx, module in enumerate(model.module_list) if isinstance(module, YOLOLayer)]                                                                                                                      
        freeze_layer_indices = [x for x in range(len(model.module_list)) if                                                                                                         
                                (x not in output_layer_indices) and                                                                                                               
                                (x - 1 not in output_layer_indices)]                                                                                                                 
        for idx in freeze_layer_indices:                                                                                                                                             
            for parameter in model.module_list[idx].parameters():                                                                                                                    
                parameter.requires_grad_(False)                                                                                                                                      


    # Mixed precision training https://github.com/NVIDIA/apex
    if mixed_precision:
        model, optimizer = amp.initialize(model, optimizer, opt_level='O1', verbosity=0)

    # Scheduler https://arxiv.org/pdf/1812.01187.pdf
    lf = lambda x: (((1 + math.cos(x * math.pi / epochs)) / 2) ** 1.0) * 0.95 + 0.05  # cosine
    scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lf)
    scheduler.last_epoch = start_epoch - 1  # see link below
    # https://discuss.pytorch.org/t/a-problem-occured-when-resuming-an-optimizer/28822

    # Plot lr schedule
    # y = []
    # for _ in range(epochs):
    #     scheduler.step()
    #     y.append(optimizer.param_groups[0]['lr'])
    # plt.plot(y, '.-', label='LambdaLR')
    # plt.xlabel('epoch')
    # plt.ylabel('LR')
    # plt.tight_layout()
    # plt.savefig('LR.png', dpi=300)

    # Dataset
    dataset = LoadImagesAndLabels(train_path, img_size, batch_size,
                                  augment=True,
                                  hyp=hyp,  # augmentation hyperparameters
                                  rect=opt.rect,  # rectangular training
                                  cache_images=opt.cache_images,
                                  single_cls=opt.single_cls)

    # Dataloader
    batch_size = min(batch_size, len(dataset))
    nw = min([os.cpu_count(), batch_size if batch_size > 1 else 0, 8])  # number of workers
    dataloader = torch.utils.data.DataLoader(dataset,
                                             batch_size=batch_size,
                                             num_workers=nw,
                                             shuffle=not opt.rect,
                                             # Shuffle=True unless rectangular training is used
                                             pin_memory=True,
                                             collate_fn=dataset.collate_fn)

    # Testloader
    testloader = torch.utils.data.DataLoader(LoadImagesAndLabels(test_path, imgsz_test, batch_size,
                                                                 hyp=hyp,
                                                                 rect=True,
                                                                 cache_images=opt.cache_images,
                                                                 single_cls=opt.single_cls),
                                             batch_size=batch_size,
                                             num_workers=nw,
                                             pin_memory=True,
                                             collate_fn=dataset.collate_fn)

    initial_rho = opt.rho
    t0 = time.time()
    """====================="""
    """ multi-rho admm train"""
    """====================="""
    if opt.admm:
        opt.notest = True
        # possible weights are '*.pt', 'yolov3-spp.pt', 'yolov3-tiny.pt' etc.
        chkpt = torch.load(weights, map_location=device)

        # load model
        try:
            # chkpt['model'] = {k: v for k, v in chkpt['model'].items() if model.state_dict()[k].numel() == v.numel()}
            model.load_state_dict(chkpt['model'], strict=False)
        except Exception as e:
            s = "%s is not compatible with %s. Specify --weights '' or specify a --cfg compatible with %s. " \
                "See https://github.com/ultralytics/yolov3/issues/657" % (opt.weights, opt.cfg, opt.weights)
            print(e)
            raise KeyError(s) from e

        del chkpt

        # Initialize distributed training
        if device.type != 'cpu' and torch.cuda.device_count() > 1 and torch.distributed.is_available():
            dist.init_process_group(backend='nccl',  # 'distributed backend'
                                    init_method='tcp://127.0.0.1:9999',  # distributed training init method
                                    world_size=1,  # number of nodes for distributed training
                                    rank=0)  # distributed training node rank
            model = torch.nn.parallel.DistributedDataParallel(model, find_unused_parameters=True)
            model.yolo_layers = model.module.yolo_layers  # move yolo layer indices to top level


        # Model parameters
        model.nc = nc  # attach number of classes to model
        model.hyp = hyp  # attach hyperparameters to model
        model.gr = 1.0  # giou loss ratio (obj_loss = 1.0 or giou)
        model.class_weights = labels_to_class_weights(dataset.labels, nc).to(device)  # attach class weights

        # Model EMA
        ema = torch_utils.ModelEMA(model)

        # Start training
        nb = len(dataloader)  # number of batches
        n_burn = max(int(0.7 * nb), 500)  # burn-in iterations, max(0.7 epochs, 500 iterations)
        maps = np.zeros(nc)  # mAP per class
        # torch.autograd.set_detect_anomaly(True)
        results = (0, 0, 0, 0, 0, 0, 0)  # 'P', 'R', 'mAP', 'F1', 'val GIoU', 'val Objectness', 'val Classification'

        print('Image sizes %g - %g train, %g test' % (imgsz_min, imgsz_max, imgsz_test))
        print('Using %g dataloader workers' % nw)
        print('Starting training for %g epochs...' % epochs)




        for i in range(opt.rho_num):
            current_rho = initial_rho * 10 ** i
            ADMM = admm.ADMM(model, file_name=PRUNE_CONFIG_PATH , rho=current_rho)
            admm.admm_initialization(opt, ADMM=ADMM, model=model)  # intialize Z variable

            for epoch in range(start_epoch, epochs):  # epoch ------------------------------------------------------------------
                print("current rho: {}".format(current_rho))

                model.train()
                masks = {}
                if opt.masked_retrain and not opt.combine_progressive:
                    print("full acc re-train masking")

                    for name, W in (model.module.named_parameters() if type(
                            model) is torch.nn.parallel.DistributedDataParallel else model.named_parameters()):
                        if name not in ADMM.prune_ratios:
                            continue
                        above_threshold, W = admm.weight_pruning(opt, W, ADMM.prune_ratios[name])
                        W.data = W
                        masks[name] = above_threshold
                elif opt.combine_progressive:
                    print("progressive admm-train/re-train masking")
                    for name, W in (model.module.named_parameters() if type(
                            model) is torch.nn.parallel.DistributedDataParallel else model.named_parameters()):
                        weight = W.cpu().detach().numpy()
                        non_zeros = weight != 0
                        non_zeros = non_zeros.astype(np.float32)
                        zero_mask = torch.from_numpy(non_zeros).cuda()
                        W = torch.from_numpy(weight).cuda()
                        W.data = W
                        masks[name] = zero_mask

                # Update image weights (optional)
                if dataset.image_weights:
                    w = model.class_weights.cpu().numpy() * (1 - maps) ** 2  # class weights
                    image_weights = labels_to_image_weights(dataset.labels, nc=nc, class_weights=w)
                    dataset.indices = random.choices(range(dataset.n), weights=image_weights, k=dataset.n)  # rand weighted idx

                mloss = torch.zeros(4).to(device)  # mean losses
                print(('\n' + '%10s' * 8) % ('Epoch', 'gpu_mem', 'GIoU', 'obj', 'cls', 'total', 'targets', 'img_size'))
                pbar = tqdm(enumerate(dataloader), total=nb)  # progress bar
                for i, (imgs, targets, paths, _) in pbar:  # batch -------------------------------------------------------------
                    if i >2:
                        break
                    ni = i + nb * epoch  # number integrated batches (since train start)
                    imgs = imgs.to(device).float() / 255.0  # uint8 to float32, 0 - 255 to 0.0 - 1.0
                    targets = targets.to(device)

                    # Burn-in
                    if ni <= n_burn:
                        xi = [0, n_burn]  # x interp
                        model.gr = np.interp(ni, xi, [0.0, 1.0])  # giou loss ratio (obj_loss = 1.0 or giou)
                        accumulate = max(1, np.interp(ni, xi, [1, 64 / batch_size]).round())
                        for j, x in enumerate(optimizer.param_groups):
                            # bias lr falls from 0.1 to lr0, all other lrs rise from 0.0 to lr0
                            x['lr'] = np.interp(ni, xi, [0.1 if j == 2 else 0.0, x['initial_lr'] * lf(epoch)])
                            x['weight_decay'] = np.interp(ni, xi, [0.0, hyp['weight_decay'] if j == 1 else 0.0])
                            if 'momentum' in x:
                                x['momentum'] = np.interp(ni, xi, [0.9, hyp['momentum']])

                    # Multi-Scale
                    if opt.multi_scale:
                        if ni / accumulate % 1 == 0:  #  adjust img_size (67% - 150%) every 1 batch
                            img_size = random.randrange(grid_min, grid_max + 1) * gs
                        sf = img_size / max(imgs.shape[2:])  # scale factor
                        if sf != 1:
                            ns = [math.ceil(x * sf / gs) * gs for x in imgs.shape[2:]]  # new shape (stretched to 32-multiple)
                            imgs = F.interpolate(imgs, size=ns, mode='bilinear', align_corners=False)

                    # Forward
                    pred = model(imgs)

                    # Loss
                    loss, loss_items = compute_loss(pred, targets, model)
                    if not torch.isfinite(loss):
                        print('WARNING: non-finite loss, ending training ', loss_items)
                        return results

                    # Backward
                    loss *= batch_size / 64  # scale loss


                    admm.z_u_update(opt, ADMM, model, device, dataloader, optimizer, epoch, imgs, i,
                                        tb_writer)  # update Z and U variables
                    loss, admm_loss, mixed_loss = admm.append_admm_loss(opt, ADMM, model,
                                                                            loss)  # append admm losss

                    if mixed_precision:
                        with amp.scale_loss(mixed_loss, optimizer) as scaled_loss:
                            scaled_loss.backward()
                    else:
                        mixed_loss.backward()

                    if opt.combine_progressive:
                        with torch.no_grad():
                            for name, W in (model.module.named_parameters() if type(
                                    model) is torch.nn.parallel.DistributedDataParallel else model.named_parameters()):
                                if name in masks:
                                    W.grad *= masks[name]

                    # Optimize
                    if ni % accumulate == 0:
                        optimizer.step()
                        optimizer.zero_grad()
                        ema.update(model)

                    # Print
                    mloss = (mloss * i + loss_items) / (i + 1)  # update mean losses
                    mem = '%.3gG' % (torch.cuda.memory_cached() / 1E9 if torch.cuda.is_available() else 0)  # (GB)
                    s = ('%10s' * 2 + '%10.3g' * 6) % ('%g/%g' % (epoch, epochs - 1), mem, *mloss, len(targets), img_size)
                    pbar.set_description(s)

                    # Plot
                    # if ni < 1:
                    #     f = 'train_batch%g.jpg' % i  # filename
                        # res = plot_images(images=imgs, targets=targets, paths=paths, fname=f)
                        # if tb_writer:
                        #     tb_writer.add_image(f, res, dataformats='HWC', global_step=epoch)
                        #     # tb_writer.add_graph(model, imgs)  # add model to tensorboard

                    # end batch ------------------------------------------------------------------------------------------------

                # Update scheduler
                if opt.admm:
                    admm.admm_adjust_learning_rate(optimizer, epoch, opt)
                else:
                    scheduler.step()

                # Process epoch results
                ema.update_attr(model)
                final_epoch = epoch + 1 == epochs
                if not opt.notest:  # Calculate mAP  #or final_epoch
                    is_coco = any([x in data for x in ['coco.data', 'coco2014.data', 'coco2017.data']]) and model.nc == 80
                    results, maps = test.test(cfg,
                                              data,
                                              batch_size=batch_size,
                                              imgsz=imgsz_test,
                                              model=ema.ema,
                                              save_json=final_epoch and is_coco,
                                              single_cls=opt.single_cls,
                                              dataloader=testloader,
                                              multi_label=ni > n_burn)

                # Write
                with open(results_file, 'a') as f:
                    f.write(s + '%10.3g' * 7 % results + '\n')  # P, R, mAP, F1, test_losses=(GIoU, obj, cls)
                if len(opt.name) and opt.bucket:
                    os.system('gsutil cp results.txt gs://%s/results/results%s.txt' % (opt.bucket, opt.name))

                # Tensorboard
                if tb_writer:
                    tags = ['train/giou_loss', 'train/obj_loss', 'train/cls_loss',
                            'metrics/precision', 'metrics/recall', 'metrics/mAP_0.5', 'metrics/F1',
                            'val/giou_loss', 'val/obj_loss', 'val/cls_loss']
                    for x, tag in zip(list(mloss[:-1]) + list(results), tags):
                        tb_writer.add_scalar(tag, x, epoch)

                # Update best mAP
                fi = fitness(np.array(results).reshape(1, -1))  # fitness_i = weighted combination of [P, R, mAP, F1]
                if fi > best_fitness:
                    best_fitness = fi

                # end epoch ----------------------------------------------------------------------------------------------------
            # end training

            # admm_adjust_learning_rate ----------------------------------------------------------------------------------------------------
            admm.admm_adjust_learning_rate(optimizer, epoch, opt)
            # end admm_adjust_learning_rate ----------------------------------------------------------------------------------------------------

            print("Saving model.")
            torch.save(
                model.module.state_dict() if type(model) is nn.parallel.DistributedDataParallel else model.state_dict(),
                "./weights/yolov4_{}_{}_{}.pt".format(
                    current_rho, opt.config_file, opt.sparsity_type))

        if not opt.evolve:
            plot_results()  # save as results.png
        print('%g epochs completed in %.3f hours.\n' % (epoch - start_epoch + 1, (time.time() - t0) / 3600))
        # dist.destroy_process_group() if torch.cuda.device_count() > 1 else None
        # torch.cuda.empty_cache()
        # return results


    """=============="""
    """masked retrain"""
    """=============="""
    if opt.masked_retrain:
        ADMM = admm.ADMM(model, file_name=PRUNE_CONFIG_PATH, rho=initial_rho)
        if not opt.resume:
            # possible weights are '*.pt', 'yolov3-spp.pt', 'yolov3-tiny.pt' etc.
            print("\n>_ Loading file: {}yolov4_{}_{}_{}.pt".format(wdir, initial_rho * 10 ** (opt.rho_num - 1), opt.config_file, opt.sparsity_type))
            chkpt = torch.load("{}yolov4_{}_{}_{}.pt".format(wdir, initial_rho * 10 ** (opt.rho_num - 1), opt.config_file, opt.sparsity_type), map_location=device)
            # chkpt = torch.load(weights, map_location=device)
            # load model
            try:
                # chkpt['model'] = {k: v for k, v in chkpt['model'].items() if model.state_dict()[k].numel() == v.numel()}
                model.load_state_dict(chkpt, strict=False) #['model']

            except KeyError as e:
                # s = "%s is not compatible with %s. Specify --weights '' or specify a --cfg compatible with %s. " \
                #     "See https://github.com/ultralytics/yolov3/issues/657" % (opt.weights, opt.cfg, opt.weights)
                raise KeyError() from e
            #----------------------------------------------hard prune------------------------------------------------
            admm.hard_prune(opt, ADMM, model)
            #----------------------------------------------hard prune------------------------------------------------
        else:
            try:
                chkpt = torch.load(weights, map_location=device)
                chkpt['model'] = {k: v for k, v in chkpt['model'].items() if model.state_dict()[k].numel() == v.numel()}
                model.load_state_dict(chkpt['model'], strict=False)
            except KeyError as e:
                # s = "%s is not compatible with %s. Specify --weights '' or specify a --cfg compatible with %s. " \
                #     "See https://github.com/ultralytics/yolov3/issues/657" % (opt.weights, opt.cfg, opt.weights)
                raise KeyError() from e
            # load optimizer
            if chkpt['optimizer'] is not None:
                optimizer.load_state_dict(chkpt['optimizer'])
                best_fitness = chkpt['best_fitness']

            # load results
            if chkpt.get('training_results') is not None:
                with open(results_file, 'w') as file:
                    file.write(chkpt['training_results'])  # write results.txt

            start_epoch = chkpt['epoch'] + 1
        del chkpt

        # Initialize distributed training
        if device.type != 'cpu' and torch.cuda.device_count() > 1 and torch.distributed.is_available():
            dist.init_process_group(backend='nccl',  # 'distributed backend'
                                    init_method='tcp://127.0.0.1:9919',  # distributed training init method
                                    world_size=1,  # number of nodes for distributed training
                                    rank=0)  # distributed training node rank
            model = torch.nn.parallel.DistributedDataParallel(model, find_unused_parameters=True)
            model.yolo_layers = model.module.yolo_layers  # move yolo layer indices to top level

            # Model parameters
        model.nc = nc  # attach number of classes to model
        model.hyp = hyp  # attach hyperparameters to model
        model.gr = 1.0  # giou loss ratio (obj_loss = 1.0 or giou)
        model.class_weights = labels_to_class_weights(dataset.labels, nc).to(device)  # attach class weights

        # Model EMA
        ema = torch_utils.ModelEMA(model)

        # Start training
        nb = len(dataloader)  # number of batches
        n_burn = max(3 * nb, 500)  # burn-in iterations, max(3 epochs, 500 iterations)
        maps = np.zeros(nc)  # mAP per class
        # torch.autograd.set_detect_anomaly(True)
        results = (0, 0, 0, 0, 0, 0, 0)  # 'P', 'R', 'mAP', 'F1', 'val GIoU', 'val Objectness', 'val Classification'
        print('Image sizes %g - %g train, %g test' % (imgsz_min, imgsz_max, imgsz_test))
        print('Using %g dataloader workers' % nw)
        print('Starting training for %g epochs...' % epochs)
        for epoch in range(start_epoch, epochs):  # epoch ------------------------------------------------------------------
            model.train()

            if opt.masked_retrain and not opt.combine_progressive:
                print("full acc re-train masking")
                masks = {}
                for name, W in (model.module.named_parameters() if type(
                        model) is torch.nn.parallel.DistributedDataParallel else model.named_parameters()):
                    if name not in ADMM.prune_ratios:
                        continue
                    above_threshold, W = admm.weight_pruning(opt, W, ADMM.prune_ratios[name])
                    W.data = W
                    masks[name] = above_threshold
            elif opt.combine_progressive:
                print("progressive admm-train/re-train masking")
                masks = {}
                for name, W in (model.module.named_parameters() if type(
                        model) is torch.nn.parallel.DistributedDataParallel else model.named_parameters()):
                    weight = W.cpu().detach().numpy()
                    non_zeros = weight != 0
                    non_zeros = non_zeros.astype(np.float32)
                    zero_mask = torch.from_numpy(non_zeros).cuda()
                    W = torch.from_numpy(weight).cuda()
                    W.data = W
                    masks[name] = zero_mask

            # Update image weights (optional)
            if dataset.image_weights:
                w = model.class_weights.cpu().numpy() * (1 - maps) ** 2  # class weights
                image_weights = labels_to_image_weights(dataset.labels, nc=nc, class_weights=w)
                dataset.indices = random.choices(range(dataset.n), weights=image_weights,
                                                 k=dataset.n)  # rand weighted idx

            mloss = torch.zeros(4).to(device)  # mean losses
            print(('\n' + '%10s' * 8) % ('Epoch', 'gpu_mem', 'GIoU', 'obj', 'cls', 'total', 'targets', 'img_size'))
            pbar = tqdm(enumerate(dataloader), total=nb)  # progress bar
            for i, (imgs, targets, paths, _) in pbar:  # batch -------------------------------------------------------------
                if i >2:
                    break
                ni = i + nb * epoch  # number integrated batches (since train start)
                imgs = imgs.to(device).float() / 255.0  # uint8 to float32, 0 - 255 to 0.0 - 1.0
                targets = targets.to(device)

                # Burn-in
                if ni <= n_burn:
                    xi = [0, n_burn]  # x interp
                    model.gr = np.interp(ni, xi, [0.0, 1.0])  # giou loss ratio (obj_loss = 1.0 or giou)
                    accumulate = max(1, np.interp(ni, xi, [1, 64 / batch_size]).round())
                    for j, x in enumerate(optimizer.param_groups):
                        # bias lr falls from 0.1 to lr0, all other lrs rise from 0.0 to lr0
                        x['lr'] = np.interp(ni, xi, [0.1 if j == 2 else 0.0, x['initial_lr'] * lf(epoch)])
                        x['weight_decay'] = np.interp(ni, xi, [0.0, hyp['weight_decay'] if j == 1 else 0.0])
                        if 'momentum' in x:
                            x['momentum'] = np.interp(ni, xi, [0.9, hyp['momentum']])

                # Multi-Scale
                if opt.multi_scale:
                    if ni / accumulate % 1 == 0:  # adjust img_size (67% - 150%) every 1 batch
                        img_size = random.randrange(grid_min, grid_max + 1) * gs
                    sf = img_size / max(imgs.shape[2:])  # scale factor
                    if sf != 1:
                        ns = [math.ceil(x * sf / gs) * gs for x in
                              imgs.shape[2:]]  # new shape (stretched to 32-multiple)
                        imgs = F.interpolate(imgs, size=ns, mode='bilinear', align_corners=False)

                # Forward
                pred = model(imgs)

                # Loss
                loss, loss_items = compute_loss(pred, targets, model)
                if not torch.isfinite(loss):
                    print('WARNING: non-finite loss, ending training ', loss_items)
                    return results

                # Backward
                loss *= batch_size / 64  # scale loss
                if mixed_precision:
                    with amp.scale_loss(loss, optimizer) as scaled_loss:
                        scaled_loss.backward()
                else:
                    loss.backward()

                if opt.combine_progressive:
                    with torch.no_grad():
                        for name, W in (model.module.named_parameters() if type(
                                model) is torch.nn.parallel.DistributedDataParallel else model.named_parameters()):
                            if name in masks:
                                W.grad *= masks[name]
                if opt.masked_retrain:
                    with torch.no_grad():
                        for name, W in (model.module.named_parameters() if type(
                                model) is torch.nn.parallel.DistributedDataParallel else model.named_parameters()):
                            if name in masks:
                                W.grad *= masks[name]

                # Optimize
                if ni % accumulate == 0:
                    optimizer.step()
                    optimizer.zero_grad()
                    ema.update(model)

                # Print
                mloss = (mloss * i + loss_items) / (i + 1)  # update mean losses
                mem = '%.3gG' % (torch.cuda.memory_cached() / 1E9 if torch.cuda.is_available() else 0)  # (GB)
                s = ('%10s' * 2 + '%10.3g' * 6) % (
                '%g/%g' % (epoch, epochs - 1), mem, *mloss, len(targets), img_size)
                pbar.set_description(s)

                # Plot
                if ni < 1:
                    f = 'train_batch%g.jpg' % i  # filename
                    res = plot_images(images=imgs, targets=targets, paths=paths, fname=f)
                    if tb_writer:
                        tb_writer.add_image(f, res, dataformats='HWC', global_step=epoch)
                        # tb_writer.add_graph(model, imgs)  # add model to tensorboard

                # end batch ------------------------------------------------------------------------------------------------

            # Update scheduler
            scheduler.step()

            # Process epoch results
            ema.update_attr(model)
            final_epoch = epoch + 1 == epochs
            if not opt.notest or final_epoch:  # Calculate mAP
                is_coco = any(
                    [x in data for x in ['coco.data', 'coco2014.data', 'coco2017.data']]) and model.nc == 80
                results, maps = test.test(cfg,
                                          data,
                                          batch_size=batch_size,
                                          imgsz=imgsz_test,
                                          model=ema.ema,
                                          save_json=final_epoch and is_coco,
                                          single_cls=opt.single_cls,
                                          dataloader=testloader,
                                          multi_label=ni > n_burn)

            # Write
            with open(results_file, 'a') as f:
                f.write(s + '%10.3g' * 7 % results + '\n')  # P, R, mAP, F1, test_losses=(GIoU, obj, cls)
            if len(opt.name) and opt.bucket:
                os.system('gsutil cp results.txt gs://%s/results/results%s.txt' % (opt.bucket, opt.name))

            # Tensorboard
            if tb_writer:
                tags = ['train/giou_loss', 'train/obj_loss', 'train/cls_loss',
                        'metrics/precision', 'metrics/recall', 'metrics/mAP_0.5', 'metrics/F1',
                        'val/giou_loss', 'val/obj_loss', 'val/cls_loss']
                for x, tag in zip(list(mloss[:-1]) + list(results), tags):
                    tb_writer.add_scalar(tag, x, epoch)

            # Update best mAP
            fi = fitness(np.array(results).reshape(1, -1))  # fitness_i = weighted combination of [P, R, mAP, F1]
            if fi > best_fitness:  #results[2]
                best_fitness = fi  #results[2]
                print("\n>_ Got better accuracy {:.3f}% now...\n".format(results[2]))
                # torch.save(ema.ema.module.state_dict() if hasattr(model, 'module') else ema.ema.state_dict(),
                #            "./model_retrained/yolov4_retrained_acc_{:.3f}_{}rhos_{}_{}.pt".format(results[2], opt.rho_num, opt.config_file, opt.sparsity_type))

            # Save model
            save = (not opt.nosave) or (final_epoch and not opt.evolve)
            if save:
                with open(results_file, 'r') as f:  # create checkpoint
                    chkpt = {'epoch': epoch,
                             'best_fitness': best_fitness,
                             'training_results': f.read(),
                             'model': ema.ema.module.state_dict() if hasattr(model,
                                                                             'module') else ema.ema.state_dict(),
                             'optimizer': None if final_epoch else optimizer.state_dict()}

                # Save last, best and delete
                torch.save(chkpt, last)
                if (best_fitness == fi) and not final_epoch:
                    torch.save(chkpt, best)
                del chkpt

            # end epoch ----------------------------------------------------------------------------------------------------
        # end training

        test_sparsity(model)
        print("Best Acc: {:.4f}".format(results[2]))
        n = opt.name
        if len(n):
            n = '_' + n if not n.isnumeric() else n
            fresults, flast, fbest = 'results%s.txt' % n, wdir + 'last%s.pt' % n, wdir + 'best%s.pt' % n
            for f1, f2 in zip([wdir + 'last.pt', wdir + 'best.pt', 'results.txt'], [flast, fbest, fresults]):
                if os.path.exists(f1):
                    os.rename(f1, f2)  # rename
                    ispt = f2.endswith('.pt')  # is *.pt
                    strip_optimizer(f2) if ispt else None  # strip optimizer
                    os.system('gsutil cp %s gs://%s/weights' % (
                    f2, opt.bucket)) if opt.bucket and ispt else None  # upload

        if not opt.evolve:
            plot_results()  # save as results.png
        print('%g epochs completed in %.3f hours.\n' % (epoch - start_epoch + 1, (time.time() - t0) / 3600))
        # dist.destroy_process_group() if torch.cuda.device_count() > 1 else None
        # torch.cuda.empty_cache()
        return results

def main_autorun(arg_list):
    global mixed_precision
    mixed_precision = True
    try:  # Mixed precision training https://github.com/NVIDIA/apex
        from apex import amp
    except:
        print('Apex recommended for faster mixed precision training: https://github.com/NVIDIA/apex')
        mixed_precision = False  # not installed


    # Hyperparameters
    hyp = {'giou': 3.54,  # giou loss gain
        'cls': 37.4,  # cls loss gain
        'cls_pw': 1.0,  # cls BCELoss positive_weight
        'obj': 64.3,  # obj loss gain (*=img_size/320 if img_size != 320)
        'obj_pw': 1.0,  # obj BCELoss positive_weight
        'iou_t': 0.20,  # iou training threshold
        'lr0': 5E-3,  # initial learning rate (SGD=5E-3, Adam=5E-4)
        'lrf': 0.0005,  # final learning rate (with cos scheduler)
        'momentum': 0.937,  # SGD momentum
        'weight_decay': 0.0005,  # optimizer weight decay
        'fl_gamma': 0.0,  # focal loss gamma (efficientDet default is gamma=1.5)
        'hsv_h': 0.0138,  # image HSV-Hue augmentation (fraction)
        'hsv_s': 0.678,  # image HSV-Saturation augmentation (fraction)
        'hsv_v': 0.36,  # image HSV-Value augmentation (fraction)
        'degrees': 1.98 * 0,  # image rotation (+/- deg)
        'translate': 0.05 * 0,  # image translation (+/- fraction)
        'scale': 0.05 * 0,  # image scale (+/- gain)
        'shear': 0.641 * 0}  # image shear (+/- deg)

    # Overwrite hyp with hyp*.txt (optional)
    f = glob.glob('hyp*.txt')
    if f:
        print('Using %s' % f[0])
        for k, v in zip(hyp.keys(), np.loadtxt(f[0])):
            hyp[k] = v

    # Print focal loss if gamma > 0
    if hyp['fl_gamma']:
        print('Using FocalLoss(gamma=%g)' % hyp['fl_gamma'])
    parser = argparse.ArgumentParser()
    parser.add_argument('--epochs', type=int, default=280)  # 500200 batches at bs 16, 117263 COCO images = 273 epochs
    parser.add_argument('--batch-size', type=int, default=64)  # effective bs = batch_size * accumulate = 16 * 4 = 64
    parser.add_argument('--cfg', type=str, default='cfg/csdarknet53s-panet-spp.cfg', help='*.cfg path')
    parser.add_argument('--data', type=str, default='data/coco2014.data', help='*.data path')
    parser.add_argument('--multi-scale', action='store_true', help='adjust (67%% - 150%%) img_size every 10 batches')
    parser.add_argument('--img-size', nargs='+', type=int, default=[256,416,320], help='[min_train, max-train, test]')
    parser.add_argument('--rect', action='store_true', help='rectangular training')
    parser.add_argument('--resume', action='store_true', help='resume training from last.pt')
    parser.add_argument('--nosave', action='store_true', help='only save final checkpoint')
    parser.add_argument('--notest', action='store_true', help='only test final epoch')
    parser.add_argument('--evolve', action='store_true', help='evolve hyperparameters')
    parser.add_argument('--bucket', type=str, default='', help='gsutil bucket')
    parser.add_argument('--cache-images', action='store_true', help='cache images for faster training')
    parser.add_argument('--weights', type=str, default='weights/yolov4dense.pt', help='initial weights path')
    parser.add_argument('--name', default='', help='renames results.txt to results_name.txt if supplied')
    parser.add_argument('--device', default='', help='device id (i.e. 0 or 0,1 or cpu)')
    parser.add_argument('--adam', action='store_true', help='use adam optimizer')
    parser.add_argument('--single-cls', action='store_true', help='train as single-class dataset')
    parser.add_argument('--freeze-layers', action='store_true', help='Freeze non-output layers')  
    parser.add_argument('--verbose', action='store_true', help='Freeze non-output layers')  

    parser.add_argument('--admm', action='store_true', help='do admm process')
    parser.add_argument('--masked-retrain', action='store_true', help='do masked-retrain process')
    parser.add_argument('--rho', default=0.0001, help='rho val')
    parser.add_argument('--rho-num', default=1, help='order of magnitude （from 0.0001 x 1e-1 to 0.0001 x 1e-4)')
    parser.add_argument('--admm-epochs', default=6, help='rho val')
    parser.add_argument('--sparsity-type', type=str, default='block-punched', help='order of magnitude （from 0.0001 x 1e-1 to 0.0001 x 1e-4)')
    parser.add_argument('--config-file', type=str, default='config_csdarknet53pan_v14', help='prune ratio config file')
    parser.add_argument('--combine-progressive', action='store_true', help='progressive pruning')

    # parser.add_argument('--admm-file', type=str, default='admm', help='admm configuration file')


    global opt
    opt = parser.parse_args(arg_list)
    
    opt.lr = hyp['lr0']
    # opt.mixup = True
    # opt.alpha = 0.3
    # opt.smooth = True
    # opt.smooth_eps = 0.1


    opt.weights = last if opt.resume and not opt.weights else opt.weights

    # check_git_status()
    # opt.cfg = check_file(opt.cfg)  # check file
    # opt.data = check_file(opt.data)  # check file
    print(opt)
    opt.img_size.extend([opt.img_size[-1]] * (3 - len(opt.img_size)))  # extend to 3 sizes (min, max, test)
    global device
    device = torch_utils.select_device(opt.device, apex=mixed_precision, batch_size=opt.batch_size)
    if device.type == 'cpu':
        mixed_precision = False

    global tb_writer
    tb_writer = None
    # print('Start Tensorboard with "tensorboard --logdir=runs", view at http://localhost:6006/')
    # tb_writer = SummaryWriter(comment=opt.name)
    train(hyp)  # train normally
    # exit(0)
if __name__ == '__main__':
    global mixed_precision

    mixed_precision = True
    try:  # Mixed precision training https://github.com/NVIDIA/apex
        from apex import amp
    except:
        print('Apex recommended for faster mixed precision training: https://github.com/NVIDIA/apex')
        mixed_precision = False  # not installed



    # Hyperparameters
    hyp = {'giou': 3.54,  # giou loss gain
        'cls': 37.4,  # cls loss gain
        'cls_pw': 1.0,  # cls BCELoss positive_weight
        'obj': 64.3,  # obj loss gain (*=img_size/320 if img_size != 320)
        'obj_pw': 1.0,  # obj BCELoss positive_weight
        'iou_t': 0.20,  # iou training threshold
        'lr0': 5E-3,  # initial learning rate (SGD=5E-3, Adam=5E-4)
        'lrf': 0.0005,  # final learning rate (with cos scheduler)
        'momentum': 0.937,  # SGD momentum
        'weight_decay': 0.0005,  # optimizer weight decay
        'fl_gamma': 0.0,  # focal loss gamma (efficientDet default is gamma=1.5)
        'hsv_h': 0.0138,  # image HSV-Hue augmentation (fraction)
        'hsv_s': 0.678,  # image HSV-Saturation augmentation (fraction)
        'hsv_v': 0.36,  # image HSV-Value augmentation (fraction)
        'degrees': 1.98 * 0,  # image rotation (+/- deg)
        'translate': 0.05 * 0,  # image translation (+/- fraction)
        'scale': 0.05 * 0,  # image scale (+/- gain)
        'shear': 0.641 * 0}  # image shear (+/- deg)

    # Overwrite hyp with hyp*.txt (optional)
    f = glob.glob('hyp*.txt')
    if f:
        print('Using %s' % f[0])
        for k, v in zip(hyp.keys(), np.loadtxt(f[0])):
            hyp[k] = v

    # Print focal loss if gamma > 0
    if hyp['fl_gamma']:
        print('Using FocalLoss(gamma=%g)' % hyp['fl_gamma'])

    parser = argparse.ArgumentParser()
    parser.add_argument('--epochs', type=int, default=280)  # 500200 batches at bs 16, 117263 COCO images = 273 epochs
    parser.add_argument('--batch-size', type=int, default=64)  # effective bs = batch_size * accumulate = 16 * 4 = 64
    parser.add_argument('--cfg', type=str, default='cfg/csdarknet53s-panet-spp.cfg', help='*.cfg path')
    parser.add_argument('--data', type=str, default='data/coco2014.data', help='*.data path')
    parser.add_argument('--multi-scale', action='store_true', help='adjust (67%% - 150%%) img_size every 10 batches')
    parser.add_argument('--img-size', nargs='+', type=int, default=[256,416,320], help='[min_train, max-train, test]')
    parser.add_argument('--rect', action='store_true', help='rectangular training')
    parser.add_argument('--resume', action='store_true', help='resume training from last.pt')
    parser.add_argument('--nosave', action='store_true', help='only save final checkpoint')
    parser.add_argument('--notest', action='store_true', help='only test final epoch')
    parser.add_argument('--evolve', action='store_true', help='evolve hyperparameters')
    parser.add_argument('--bucket', type=str, default='', help='gsutil bucket')
    parser.add_argument('--cache-images', action='store_true', help='cache images for faster training')
    parser.add_argument('--weights', type=str, default='weights/yolov4dense.pt', help='initial weights path')
    parser.add_argument('--name', default='', help='renames results.txt to results_name.txt if supplied')
    parser.add_argument('--device', default='', help='device id (i.e. 0 or 0,1 or cpu)')
    parser.add_argument('--adam', action='store_true', help='use adam optimizer')
    parser.add_argument('--single-cls', action='store_true', help='train as single-class dataset')
    parser.add_argument('--freeze-layers', action='store_true', help='Freeze non-output layers')  
    parser.add_argument('--verbose', action='store_true', help='Freeze non-output layers')  

    parser.add_argument('--admm', action='store_true', help='do admm process')
    parser.add_argument('--masked-retrain', action='store_true', help='do masked-retrain process')
    parser.add_argument('--rho', default=0.0001, help='rho val')
    parser.add_argument('--rho-num', default=4, help='order of magnitude （from 0.0001 x 1e-1 to 0.0001 x 1e-4)')
    parser.add_argument('--admm-epochs', default=6, help='rho val')
    parser.add_argument('--sparsity-type', type=str, default='block-punched', help='order of magnitude （from 0.0001 x 1e-1 to 0.0001 x 1e-4)')
    parser.add_argument('--config-file', type=str, default='config_csdarknet53pan_v14', help='prune ratio config file')
    parser.add_argument('--combine-progressive', action='store_true', help='progressive pruning')

    # parser.add_argument('--admm-file', type=str, default='admm', help='admm configuration file')

    global opt

    opt = parser.parse_args()
    
    opt.lr = hyp['lr0']
    # opt.mixup = True
    # opt.alpha = 0.3
    # opt.smooth = True
    # opt.smooth_eps = 0.1


    opt.weights = last if opt.resume and not opt.weights else opt.weights

    # check_git_status()
    # opt.cfg = check_file(opt.cfg)  # check file
    # opt.data = check_file(opt.data)  # check file
    print(opt)
    opt.img_size.extend([opt.img_size[-1]] * (3 - len(opt.img_size)))  # extend to 3 sizes (min, max, test)
    
    global device
    device = torch_utils.select_device(opt.device, apex=mixed_precision, batch_size=opt.batch_size)
    if device.type == 'cpu':
        mixed_precision = False

    
    # tb_writer = None
    # print('Start Tensorboard with "tensorboard --logdir=runs", view at http://localhost:6006/')
    # tb_writer = SummaryWriter(comment=opt.name)
    train(hyp)  # train normally
    # exit(0)
    


import torchvision
import random
from PIL import Image, ImageOps
import numpy as np
import numbers
import math
import torch

class GroupRandomCrop(object):
    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img_group):

        w, h = img_group[0].size
        th, tw = self.size

        out_images = list()

        x1 = random.randint(0, w - tw)
        y1 = random.randint(0, h - th)

        for img in img_group:
            assert(img.size[0] == w and img.size[1] == h)
            if w == tw and h == th:
                out_images.append(img)
            else:
                out_images.append(img.crop((x1, y1, x1 + tw, y1 + th)))

        return out_images


class GroupCenterCrop(object):
    def __init__(self, size):
        self.worker = torchvision.transforms.CenterCrop(size)

    def __call__(self, img_group):
        return [self.worker(img) for img in img_group]


class GroupRandomHorizontalFlip(object):
    """Randomly horizontally flips the given PIL.Image with a probability of 0.5
    """
    def __init__(self, is_flow=False):
        self.is_flow = is_flow

    def __call__(self, img_group, is_flow=False):
        v = random.random()
        if v < 0.5:
            ret = [img.transpose(Image.FLIP_LEFT_RIGHT) for img in img_group]
            if self.is_flow:
                for i in range(0, len(ret), 2):
                    ret[i] = ImageOps.invert(ret[i])  # invert flow pixel values when flipping
            return ret
        else:
            return img_group


class GroupNormalize(object):
    def __init__(self, mean, std, threed_data=False):
        self.threed_data = threed_data
        if self.threed_data:
            # convert to the proper format
            self.mean = torch.FloatTensor(mean).view(len(mean), 1, 1, 1)
            self.std = torch.FloatTensor(std).view(len(std), 1, 1, 1)
        else:
            self.mean = mean
            self.std = std

    def __call__(self, tensor):

        if self.threed_data:
            tensor.sub_(self.mean).div_(self.std)
        else:
            rep_mean = self.mean * (tensor.size()[0] // len(self.mean))
            rep_std = self.std * (tensor.size()[0] // len(self.std))

            # TODO: make efficient
            for t, m, s in zip(tensor, rep_mean, rep_std):
                t.sub_(m).div_(s)

        return tensor


class GroupScale(object):
    """ Rescales the input PIL.Image to the given 'size'.
    'size' will be the size of the smaller edge.
    For example, if height > width, then image will be
    rescaled to (size * height / width, size)
    size: size of the smaller edge
    interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        self.worker = torchvision.transforms.Resize(size, interpolation)

    def __call__(self, img_group):
        return [self.worker(img) for img in img_group]

class GroupRandomScale(object):
    """ Rescales the input PIL.Image to the given 'size'.
    'size' will be the size of the smaller edge.
    For example, if height > width, then image will be
    rescaled to (size * height / width, size)
    size: size of the smaller edge
    interpolation: Default: PIL.Image.BILINEAR

    Randomly select the smaller edge from the range of 'size'.
    """
    def __init__(self, size, interpolation=Image.BILINEAR):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img_group):
        selected_size = np.random.randint(low=self.size[0], high=self.size[1] + 1, dtype=int)
        scale = GroupScale(selected_size, interpolation=self.interpolation)
        return scale(img_group)

class GroupOverSample(object):
    def __init__(self, crop_size, scale_size=None, num_crops=5, flip=False):
        self.crop_size = crop_size if not isinstance(crop_size, int) else (crop_size, crop_size)

        if scale_size is not None:
            self.scale_worker = GroupScale(scale_size)
        else:
            self.scale_worker = None

        if num_crops not in [1, 3, 5, 10]:
            raise ValueError("num_crops should be in [1, 3, 5, 10] but ({})".format(num_crops))
        self.num_crops = num_crops

        self.flip = flip

    def __call__(self, img_group):

        if self.scale_worker is not None:
            img_group = self.scale_worker(img_group)

        image_w, image_h = img_group[0].size
        crop_w, crop_h = self.crop_size

        if self.num_crops == 3:
            w_step = (image_w - crop_w) // 4
            h_step = (image_h - crop_h) // 4
            offsets = list()
            if image_w < image_h:
                offsets.append((2 * w_step, 0 * h_step))  # top
                offsets.append((2 * w_step, 4 * h_step))  # bottom
                offsets.append((2 * w_step, 2 * h_step))  # center
            else:
                offsets.append((0 * w_step, 2 * h_step))  # left
                offsets.append((4 * w_step, 2 * h_step))  # right
                offsets.append((2 * w_step, 2 * h_step))  # center

        else:
            offsets = GroupMultiScaleCrop.fill_fix_offset(False, image_w, image_h, crop_w, crop_h)

        oversample_group = list()
        for o_w, o_h in offsets:
            normal_group = list()
            flip_group = list()
            for i, img in enumerate(img_group):
                crop = img.crop((o_w, o_h, o_w + crop_w, o_h + crop_h))
                normal_group.append(crop)
                if self.flip:
                    flip_crop = crop.copy().transpose(Image.FLIP_LEFT_RIGHT)

                    if img.mode == 'L' and i % 2 == 0:
                        flip_group.append(ImageOps.invert(flip_crop))
                    else:
                        flip_group.append(flip_crop)

            oversample_group.extend(normal_group)
            if self.flip:
                oversample_group.extend(flip_group)
        return oversample_group


class GroupMultiScaleCrop(object):

    def __init__(self, input_size, scales=None, max_distort=1, fix_crop=True, more_fix_crop=True):
        self.scales = scales if scales is not None else [1, 875, .75, .66]
        self.max_distort = max_distort
        self.fix_crop = fix_crop
        self.more_fix_crop = more_fix_crop
        self.input_size = input_size if not isinstance(input_size, int) else [input_size, input_size]
        self.interpolation = Image.BILINEAR

    def __call__(self, img_group):

        im_size = img_group[0].size

        crop_w, crop_h, offset_w, offset_h = self._sample_crop_size(im_size)
        crop_img_group = [img.crop((offset_w, offset_h, offset_w + crop_w, offset_h + crop_h)) for img in img_group]
        ret_img_group = [img.resize((self.input_size[0], self.input_size[1]), self.interpolation)
                         for img in crop_img_group]
        return ret_img_group

    def _sample_crop_size(self, im_size):
        image_w, image_h = im_size[0], im_size[1]

        # find a crop size
        base_size = min(image_w, image_h)
        crop_sizes = [int(base_size * x) for x in self.scales]
        crop_h = [self.input_size[1] if abs(x - self.input_size[1]) < 3 else x for x in crop_sizes]
        crop_w = [self.input_size[0] if abs(x - self.input_size[0]) < 3 else x for x in crop_sizes]

        pairs = []
        for i, h in enumerate(crop_h):
            for j, w in enumerate(crop_w):
                if abs(i - j) <= self.max_distort:
                    pairs.append((w, h))

        crop_pair = random.choice(pairs)
        if not self.fix_crop:
            w_offset = random.randint(0, image_w - crop_pair[0])
            h_offset = random.randint(0, image_h - crop_pair[1])
        else:
            w_offset, h_offset = self._sample_fix_offset(image_w, image_h, crop_pair[0], crop_pair[1])

        return crop_pair[0], crop_pair[1], w_offset, h_offset

    def _sample_fix_offset(self, image_w, image_h, crop_w, crop_h):
        offsets = self.fill_fix_offset(self.more_fix_crop, image_w, image_h, crop_w, crop_h)
        return random.choice(offsets)

    @staticmethod
    def fill_fix_offset(more_fix_crop, image_w, image_h, crop_w, crop_h):
        w_step = (image_w - crop_w) // 4
        h_step = (image_h - crop_h) // 4

        ret = list()
        ret.append((0, 0))  # upper left
        ret.append((4 * w_step, 0))  # upper right
        ret.append((0, 4 * h_step))  # lower left
        ret.append((4 * w_step, 4 * h_step))  # lower right
        ret.append((2 * w_step, 2 * h_step))  # center

        if more_fix_crop:
            ret.append((0, 2 * h_step))  # center left
            ret.append((4 * w_step, 2 * h_step))  # center right
            ret.append((2 * w_step, 4 * h_step))  # lower center
            ret.append((2 * w_step, 0 * h_step))  # upper center

            ret.append((1 * w_step, 1 * h_step))  # upper left quarter
            ret.append((3 * w_step, 1 * h_step))  # upper right quarter
            ret.append((1 * w_step, 3 * h_step))  # lower left quarter
            ret.append((3 * w_step, 3 * h_step))  # lower righ quarter

        return ret


class GroupRandomSizedCrop(object):
    """Random crop the given PIL.Image to a random size of (0.08 to 1.0) of the original size
    and and a random aspect ratio of 3/4 to 4/3 of the original aspect ratio
    This is popularly used to train the Inception networks
    size: size of the smaller edge
    interpolation: Default: PIL.Image.BILINEAR
    """
    def __init__(self, size, interpolation=Image.BILINEAR):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img_group):
        for attempt in range(10):
            area = img_group[0].size[0] * img_group[0].size[1]
            target_area = random.uniform(0.08, 1.0) * area
            aspect_ratio = random.uniform(3. / 4, 4. / 3)

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if random.random() < 0.5:
                w, h = h, w

            if w <= img_group[0].size[0] and h <= img_group[0].size[1]:
                x1 = random.randint(0, img_group[0].size[0] - w)
                y1 = random.randint(0, img_group[0].size[1] - h)
                found = True
                break
        else:
            found = False
            x1 = 0
            y1 = 0

        if found:
            out_group = list()
            for img in img_group:
                img = img.crop((x1, y1, x1 + w, y1 + h))
                assert(img.size == (w, h))
                out_group.append(img.resize((self.size, self.size), self.interpolation))
            return out_group
        else:
            # Fallback
            scale = GroupScale(self.size, interpolation=self.interpolation)
            crop = GroupRandomCrop(self.size)
            return crop(scale(img_group))


class Stack(object):

    def __init__(self, roll=False, threed_data=False):
        self.roll = roll
        self.threed_data = threed_data

    def __call__(self, img_group):
        if img_group[0].mode == 'L':
            return np.concatenate([np.expand_dims(x, 2) for x in img_group], axis=2)
        elif img_group[0].mode == 'RGB':
            if self.threed_data:
                return np.stack(img_group, axis=0)
            else:
                if self.roll:
                    return np.concatenate([np.array(x)[:, :, ::-1] for x in img_group], axis=2)
                else:
                    return np.concatenate(img_group, axis=2)


class ToTorchFormatTensor(object):
    """ Converts a PIL.Image (RGB) or numpy.ndarray (H x W x C) in the range [0, 255]
    to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0] """
    def __init__(self, div=True, num_clips_crops=1):
        self.div = div
        self.num_clips_crops = num_clips_crops

    def __call__(self, pic):
        if isinstance(pic, np.ndarray):
            # handle numpy array
            if len(pic.shape) == 4:
                # ((NF)xCxHxW) --> (Cx(NF)xHxW)
                img = torch.from_numpy(pic).permute(3, 0, 1, 2).contiguous()
            else:  # data is HW(FC)
                img = torch.from_numpy(pic).permute(2, 0, 1).contiguous()
        else:
            # handle PIL Image
            img = torch.ByteTensor(torch.ByteStorage.from_buffer(pic.tobytes()))
            img = img.view(pic.size[1], pic.size[0], len(pic.mode))
            # put it from HWC to CHW format
            # yikes, this transpose takes 80% of the loading time/CPU
            img = img.transpose(0, 1).transpose(0, 2).contiguous()
        return img.float().div(255) if self.div else img.float()


class IdentityTransform(object):

    def __call__(self, data):
        return data


if __name__ == "__main__":
    trans = torchvision.transforms.Compose([
        GroupScale(256),
        GroupRandomCrop(224),
        Stack(),
        ToTorchFormatTensor(),
        GroupNormalize(
            mean=[.485, .456, .406],
            std=[.229, .224, .225]
        )]
    )

    im = Image.open('../tensorflow-model-zoo.torch/lena_299.png')

    color_group = [im] * 3
    rst = trans(color_group)

    gray_group = [im.convert('L')] * 9
    gray_rst = trans(gray_group)

    trans2 = torchvision.transforms.Compose([
        GroupRandomSizedCrop(256),
        Stack(),
        ToTorchFormatTensor(),
        GroupNormalize(
            mean=[.485, .456, .406],
            std=[.229, .224, .225])
    ])
    print(trans2(color_group))



import os
from ModelOpt.AutoYOLO.train import main_autorun
from ModelOpt.AutoYOLO.test import test_autorun

class YOLO_Autorun:
    def print_info(self, info):
        print('---------------------------------------------------')
        print(info)
        print('---------------------------------------------------')

    
    def run(self):
        str1 = input("Please enter the accuracy (mAP 0.5) requirement (default 48):")
        str2 = input("Please enter the latency requirement (default 5.2):")
        str3 = input("Please enter the device number, split by ',' (default 0,1,2,3):")

        #acc_req = requirement['accuracy']
        #speed_req = requirement['inference_time']

        if str1 != '':
            acc_req = float(str1)
        else:
            acc_req = 48
        if str2 != '':
            speed_req = float(str2)
        else:
            speed_req = 5.2
        if str3 != '':
            device = str(str3)
        else:
            device = '0,1,2,3'
        print('The accuracy requirement is {}'.format(acc_req))
        print('The speed requirement is {}'.format(speed_req))

        if acc_req > 48 and  acc_req < 50:
            prune_file = 'config_csdarknet53pan_v14'
            sparsity = 'block-punched'
        elif acc_req >= 50 and  acc_req < 51:
            prune_file = 'config_csdarknet53pan_v10'
            sparsity = 'block-punched'
        elif acc_req >= 51 and  acc_req < 52:
            prune_file = 'config_csdarknet53pan_v8'
            sparsity = 'block-punched'
        elif acc_req >= 52:
            prune_file = 'config_csdarknet53pan_v4'
            sparsity = 'block-punched'
        else:
            prune_file = 'config_csdarknet53pan_v14'
            sparsity = 'block-punched'
            
        # ckpts_prune = 'ckpts/ckpt_a'
        # ckpts_retrain = 'ckpts/ckpt_m'
        # ckpts_final = 'ckpts/ckpt_f'
        # if not os.path.exists(ckpts_prune):
        #     os.makedirs(ckpts_prune)
        # if not os.path.exists(ckpts_retrain):
        #     os.makedirs(ckpts_retrain)
        # if not os.path.exists(ckpts_final):
        #     os.makedirs(ckpts_final)

        upperpath = 'ModelOpt/AutoYOLO/'

        cmd_prune = 'python {}train.py --admm --epoch 25 --cfg {}cfg/csdarknet53s-panet-spp.cfg --data {}data/coco2014.data --weights weights/yolov4dense.pt --device {} --config-file {} --sparsity-type {}'.format(upperpath, upperpath, upperpath, device, prune_file, sparsity)
        cmd_retrain = 'python {}train.py --masked-retrain --epoch 280 --cfg {}cfg/csdarknet53s-panet-spp.cfg --data {}data/coco2014.data --device {} --config-file {} --sparsity-type {} --multi-scale'.format(upperpath, upperpath, upperpath, device, prune_file, sparsity)

        # cmd_test_prune = 'python pytorch/test_mask.py --model_dir={}/best_checkpoints --config-file={} --sparsity-type={}'.format(ckpts_prune, prune_file, sparsity)
        cmd_test_retrain = 'python {}test.py --device {} --cfg {}cfg/csdarknet53s-panet-spp.cfg --data {}data/coco2014.data --weights {}weights/best.pt'.format(upperpath, device, upperpath, upperpath, upperpath)


        # copy_ckpts_original = 'cp ckpts/ckpts_original/* {}'.format(ckpts_prune) 
        # os.system(copy_ckpts_original)
        # print_info('obtain original model')
            
        self.print_info('start prune.')

        # print(cmd_prune)
        # os.system(cmd_prune)
        main_autorun(['--admm', 
                        '--epoch', '1', 
                        '--cfg', upperpath+'cfg/csdarknet53s-panet-spp.cfg', 
                        '--data', upperpath+'data/coco2014.data', 
                        '--weights', 'weights/yolov4dense.pt', 
                        '--device', device,
                        '--config-file', prune_file,
                        '--sparsity-type', sparsity])

        # print_info('finish admm prune process')
        # copy_ckpts_prune = 'cp {}/best_checkpoints/*  {}'.format(ckpts_prune, ckpts_retrain)
        # os.system(copy_ckpts_prune)
        # print_info('obtain pruned model')

        self.print_info('finish prune, start retrain.')

        # print(cmd_retrain)
        # os.system(cmd_retrain)
        main_autorun(['--masked-retrain', 
                        '--epoch', '1', 
                        '--cfg', upperpath+'cfg/csdarknet53s-panet-spp.cfg', 
                        '--data', upperpath+'data/coco2014.data', 
                        '--multi-scale',
                        '--device', device,
                        '--config-file', prune_file,
                        '--sparsity-type', sparsity])

        self.print_info('finish retrain.')

        # print(cmd_test_retrain)
        # os.system(cmd_test_retrain)

        test_autorun(['--device', device,
                        '--cfg', upperpath+'cfg/csdarknet53s-panet-spp.cfg',
                        '--data', upperpath+'data/coco2014.data',
                        '--weights', 'weights/best.pt'])
        self.print_info('finish retrain with the above accuracy')



        self.print_info('obtain final model')

        self.print_info('you can find the model at ./weights/best.pt')

        
    
if __name__ == "__main__":
    ac = YOLO_Autorun()
    ac.run()


import os
import yaml
# from .opt_main import main as t_admm
# from .options import parse_args as p_admm
# from .pytorch.train_mask_t import main as t_mask, parse_args as p_mask
# from .pytorch.test_mask_t import main as test_mask, parse_args as test_parse


def main():
    def print_info(info):
        print('---------------------------------------------------')
        print(info)
        print('---------------------------------------------------')

    str1 = input("Please enter the accuracy requirement (default 0.8):")
    str2 = input("Please enter the latency requirement (default none):")
    str3 = input("Please enter the data location (default '/raid10/ms/dataset'):")

    str4 = input("Please enter the architecture name (default c3d):")
    # str4 = input("Please enter the architecture name (default c3d):")

    # acc_req = requirement['accuracy']
    # speed_req = requirement['inference_time']

    if str1 != '':
        acc_req = float(str1)
    else:
        acc_req = 0.8
    if str2 != '':
        speed_req = float(str2)
    else:
        speed_req = 1000000
    if str3 != '':
        data_location = str3
    else:
        data_location = '/raid10/ms/dataset'
    if str4 != '':
        arch = str4
    else:
        arch = 'c3d'

    print('The accuracy requirement is {}'.format(acc_req))
    print('The speed requirement is {}'.format(speed_req))
    print('The data location is {}'.format(data_location))
    print('The architecture is {}'.format(arch))

    # ckpts_prune = 'ckpts/ckpt_a'
    # ckpts_retrain = 'ckpts/ckpt_m'
    ckpts_final = 'ckpts/ckpt_f'
    # if not os.path.exists(ckpts_prune):
    #     os.makedirs(ckpts_prune)
    # if not os.path.exists(ckpts_retrain):
    #     os.makedirs(ckpts_retrain)
    if not os.path.exists(ckpts_final):
        os.makedirs(ckpts_final)

    if arch == 'c3d':
        if acc_req > 0.8:
            prune_file = 'c3d_2.53x_v2'
            flops_prune_rate='2.53x'
        else:
            prune_file = 'c3d_3.57x'
            flops_prune_rate = '3.57x'
    elif arch== 'r2+1d' or 'r2+1d-pretrained':
        if acc_req >= 0.92:
            prune_file = 'r2+1d-pretrained_2.56x'
            flops_prune_rate = '2.56x'
        else:
            prune_file = 'r2+1d-pretrained_3.20x'
            flops_prune_rate = '3.20x'
    elif arch== 's3d':
        prune_file = 's3d_2.09x'
        flops_prune_rate = '2.09x'
    else:
        print('model arch not found')

    cmd_prune = 'CUDA_VISIBLE_DEVICES=0,1,2,3 python ./ModelOpt/rt3d_pruning/opt_main.py --multi-gpu --dataset ucf101 --sparsity-type blk-kgs --connectivity-block-size 8 4 --admm --rho 0.0001 --rho-num 4 --epoch 50 --lr 5e-4 --optmzr sgd --log-interval 50 --smooth --smooth-eps 0.1 --data_location=\'{}\''.format(
        data_location)

    cmd_retrain = 'CUDA_VISIBLE_DEVICES=0,1,2,3 python ./ModelOpt/rt3d_pruning/opt_main.py --multi-gpu --dataset ucf101 --sparsity-type blk-kgs --connectivity-block-size 8 4 --combine-progressive --masked-retrain --rho 0.0001 --rho-num 4 --epoch 130 --lr 5e-4 --optmzr sgd --log-interval 50 --warmup --warmup-lr 1e-5 --lr-scheduler cosine --distill --teacharch c3d --teacher-path checkpoint/ucf101_c3d_transfer_epoch-14_top1-81.571.pt #--smooth --smooth-eps 0.1 --data_location=\'{}\''.format(
        data_location)

    # cmd_test_prune = 'python pytorch/test_mask.py --model_dir=\'{}\'/best_checkpoints --data_location=\'{}\' --sparsity-type=\'{}\''.format(
    #     ckpts_prune, data_location, sparsity)
    # cmd_test_retrain = 'python pytorch/test_mask.py --model_dir=\'{}\'/best_checkpoints --data_location=\'{}\' --sparsity-type=\'{}\''.format(
    #     ckpts_retrain, data_location, sparsity)

    cmd_prune_new = cmd_prune + ' --arch {} --config-file=\'{}\''.format(arch, prune_file)  # must leave a space at the beginning
    cmd_retrain_new = cmd_retrain + ' --arch {} --config-file=\'{}\''.format(arch, prune_file)

    print_info('the estimated FLOPs pruning rate is {}.'.format(flops_prune_rate))

    print_info('start pruning')

    print(cmd_prune_new) #Todo: test package

    # parser_admm = p_admm()
    # args_admm = parser_admm.parse_args(['--data_location', data_location]) # note the square bracket
    # t_admm(args_admm)
    os.system(cmd_prune_new)



    print_info('finish prune, start retrain.')

    print(cmd_retrain_new)  #Todo: test package

    # parser_mask = t_admm()
    # args_mask = parser_mask.parse_args(['--data_location', data_location]) # note the square bracket
    # t_admm(args_mask)
    os.system(cmd_retrain_new)





    print_info('you can find the model at checkpoint folder')



# compare and obtain the best accuracy
# acc = np.zeros(4)
# for i in range(4):
#	acc[i] = open("ckpts/ckpt_m_{}/acc.txt".format(i)).read().float()

# maxlocation = np.argmax(acc)

# print_info('you can find the model at ckpts/ckpt_m_{} with accuracy {}'.format(maxlocation, acc[maxlocation]))

#######################################################################################################
#     for i in range(4):
#         print_info('start prune ratio case {}.'.format(i))
#
#         ckpts_prune_new = 'ckpts/ckpt_a_{}'.format(i)
#         ckpts_retrain_new = 'ckpts/ckpt_m_{}'.format(i)
#         ckpts_final_new = 'ckpts/ckpt_f_{}'.format(i)
#         if not os.path.exists(ckpts_prune_new):
#             os.makedirs(ckpts_prune_new)
#         if not os.path.exists(ckpts_retrain_new):
#             os.makedirs(ckpts_retrain_new)
#         if not os.path.exists(ckpts_final_new):
#             os.makedirs(ckpts_final_new)
#
#         copy_ckpts_original = 'cp ckpts/ckpts_original/* {}'.format(ckpts_prune_new)
#         os.system(copy_ckpts_original)
#         print_info('obtain original model')
#
#         print_info('start prune.')
#
#         cmd_prune_new = cmd_prune + ' --config-file=\'ratio_{}\''.format(i)
#         cmd_retrain_new = cmd_retrain + ' --config-file=\'ratio_{}\''.format(i)
#         cmd_test_prune_new = cmd_test_prune + ' --config-file=\'ratio_{}\''.format(i)
#         cmd_test_retrain_new = cmd_test_retrain + ' --config-file=\'ratio_{}\''.format(i)
#         print(cmd_prune_new)
#
#         #  parser_admm = p_admm()
#         # args_admm = parser_admm.parse_args(['--model_dir', ckpts_prune_new, '--data_location', data_location, '--sparsity-type', sparsity, '--config-file', 'ratio_{}'.format(i)]) # note the square bracket
#         #  t_admm(args_admm)
#         # os.system(cmd_prune_new)
#
#         print(cmd_test_prune_new)
#         # os.system(cmd_test_prune_new)
#         # parser_test = test_parse()
#         #   args_test = parser_test.parse_args(['--model_dir', ckpts_prune_new, '--data_location', data_location, '--sparsity-type', sparsity, '--config-file', 'ratio_{}'.format(i)]) # note the square bracket
#         #  test_mask(args_test)
#         print_info('finish prune with the above accuracy')
#
#         copy_ckpts_prune = 'cp {}/best_checkpoints/*  {}'.format(ckpts_prune_new, ckpts_retrain_new)
#         os.system(copy_ckpts_prune)
#         print_info('obtain pruned model')
#
#         print_info('finish prune, start retrain.')
#
#         print(cmd_retrain_new)
#         # os.system(cmd_retrain_new)
#         #    parser_mask = p_mask()
#         #   args_mask = parser_mask.parse_args(['--model_dir', ckpts_retrain_new, '--data_location', data_location, '--sparsity-type', sparsity, '--config-file', 'ratio_{}'.format(i)]) # note the square bracket
#         # to overwrite default, use parser.parse_args(['--input', 'foobar.txt'])
#         #  t_mask(args_mask)
#
#         print_info('finish retrain.')
#
#         print(cmd_test_retrain_new)
#         # os.system(cmd_test_retrain_new)
#         #  parser_test = test_parse()
#         #  args_test = parser_test.parse_args(['--model_dir', ckpts_retrain_new, '--data_location', data_location, '--sparsity-type', sparsity, '--config-file', 'ratio_{}'.format(i)]) # note the square bracket
#         #  test_mask(args_test)
#         print_info('finish retrain with the above accuracy')
#
#         copy_ckpts_prune = 'cp {}/best_checkpoints/*  {}'.format(ckpts_retrain_new, ckpts_final_new)
#         os.system(copy_ckpts_prune)
#         print_info('obtain final model')
#
#         print_info('you can find the model at {}'.format(ckpts_final_new))
#     print_info('you can find the model with the best accuracy by comparing the above.')
#
#
# # compare and obtain the best accuracy
# # acc = np.zeros(4)
# # for i in range(4):
# #	acc[i] = open("ckpts/ckpt_m_{}/acc.txt".format(i)).read().float()
#
# # maxlocation = np.argmax(acc)
#
# # print_info('you can find the model at ckpts/ckpt_m_{} with accuracy {}'.format(maxlocation, acc[maxlocation]))

if __name__ == '__main__':
    main()










from __future__ import print_function
import os
import torch
import torch.nn as nn
import numpy as np
from numpy import linalg as LA
import yaml
from ModelOpt.rt3d_pruning.utils import *


class ADMM:
    def __init__(self, model, file_name, rho=0.001):
        self.ADMM_U = {}
        self.ADMM_Z = {}
        self.rho = rho
        self.rhos = {}

        self.init(file_name, model)

    def init(self, config, model):
        '''
        Args:
            config: configuration file that has settings for prune ratios, rhos
        called by ADMM constructor. config should be a .yaml file

        '''
        with open(config, 'r') as stream:
            raw_dict = yaml.full_load(stream)
            if 'prune_ratios' in raw_dict:
                self.prune_cfg = raw_dict['prune_ratios']
            for k, v in self.prune_cfg.items():
                self.rhos[k] = self.rho
            for (name, W) in model.named_parameters():
                if name not in self.prune_cfg:
                    continue
                self.ADMM_U[name] = torch.zeros(W.shape).cuda()  # add U
                self.ADMM_Z[name] = torch.Tensor(W.shape).cuda()  # add Z


def weight_pruning(args, weight_in, prune_ratio):
    '''
    weight pruning [irregular,column,filter]
    Args:
         weight (pytorch tensor): weight tensor, ordered by output_channel, intput_channel, kernel width and kernel height
         prune_ratio (float between 0-1): target sparsity of weights

    Returns:
         mask for nonzero weights used for retraining
         a pytorch tensor whose elements/column/row that have lowest l2 norms(equivalent to absolute weight here) are set to zero

    '''
    weight = weight_in.clone().detach()  # gpu tensor on pytorch (original)
    percent = prune_ratio * 100

    if args.sparsity_type in ['blk-vanilla', 'blk-kgs']:
        global shape, ext_shape, blk_h, blk_w, num_blk_h, num_blk_w, kernel_size
        shape = weight.shape
        block_shape = args.connectivity_block_size

        ext_shape = [(shape[i] + block_shape[i] - 1) // block_shape[i] * block_shape[i] for i in range(2)] + list(shape[2:])

        blk_h, blk_w = block_shape
        num_blk_h, num_blk_w = ext_shape[0]//blk_h, ext_shape[1]//blk_w
        kernel_size = torch.prod(torch.tensor(shape[2:]))

        padding = nn.ZeroPad2d((0, ext_shape[1] - shape[1], 0, ext_shape[0] - shape[0]))
        weight_ = weight.reshape(shape[0], shape[1], -1)
        weight_ = torch.stack([padding(weight_[:, :, i]) for i in range(kernel_size)], dim=2)

    if args.sparsity_type == 'filter':
        shape = weight.shape
        weight2d = weight.reshape(shape[0], -1)
        row_l2_norm = torch.norm(weight2d, 2, dim=1)
        percentile = np.percentile(row_l2_norm.cpu(), percent)
        under_threshold = row_l2_norm <= percentile
        above_threshold = row_l2_norm > percentile
        weight2d[under_threshold, :] = 0
        above_threshold = above_threshold.type(torch.float32)
        return above_threshold, weight

    elif args.sparsity_type == 'blk-vanilla':
        weight_ = reshape_matrix2block(weight_, blk_h, blk_w)
        l2_norm = torch.norm(weight_, 2, dim=1)  # gpu tensor

        valid = torch.Tensor(num_blk_h * num_blk_w)
        for i in range(num_blk_h * num_blk_w):
            valid[i] = max(len((weight_[i]!=0).nonzero()), 1)
        if 'cuda' in str(l2_norm.device):
            valid = valid.cuda()
        l2_norm_mean = torch.div(l2_norm, torch.sqrt(valid))

        percentile = np.percentile(l2_norm_mean.cpu(), percent)
        under_threshold = l2_norm_mean <= percentile
        above_threshold = l2_norm_mean > percentile
        weight_[under_threshold, :] = 0
        above_threshold = above_threshold.type(torch.float32)

        weight_ = reshape_block2matrix(weight_, num_blk_h, num_blk_w, blk_h, blk_w * kernel_size)
        weight = weight_.reshape(ext_shape)[:shape[0], :shape[1]]

        return above_threshold, weight

    elif args.sparsity_type == 'blk-kgs':
        weight_ = reshape_matrix2block_kernel(weight_, blk_h, blk_w)
        l2_norm = torch.norm(weight_, 2, dim=1)  # gpu tensor

        valid = torch.Tensor(num_blk_h * num_blk_w * int(kernel_size))
        for i in range(num_blk_h * num_blk_w * kernel_size):
            valid[i] = max(len((weight_[i]!=0).nonzero()), 1)
        if 'cuda' in str(l2_norm.device):
            valid = valid.cuda()
        l2_norm_mean = torch.div(l2_norm, torch.sqrt(valid))

        percentile = np.percentile(l2_norm_mean.cpu(), percent)
        under_threshold = l2_norm_mean <= percentile
        above_threshold = l2_norm_mean > percentile
        weight_[under_threshold, :] = 0
        above_threshold = above_threshold.type(torch.float32)

        weight_ = reshape_block2matrix_kernel(weight_, num_blk_h, num_blk_w, blk_h, blk_w, int(kernel_size))
        weight = weight_.reshape(ext_shape)[:shape[0], :shape[1]]
        return above_threshold, weight


def hard_prune(args, ADMM, model):
    '''
    hard_pruning, or direct masking
    Args:
         model: contains weight tensors in cuda

    '''
    print('hard pruning\n')

    index_npz = os.path.join(args.ckpt_dir, '{}_{}_{}_{}rhos_index.npz'.format(args.dataset, args.arch, \
        args.sparsity_type + args.blk_str, args.rho_num))
    index_dict = {}

    for (name, W) in model.named_parameters():
        if name not in ADMM.prune_cfg:  # ignore layers that do not have rho
            continue
        retained_index, cuda_pruned_weights = weight_pruning(args, W, ADMM.prune_cfg[name])  # get sparse model in cuda
        if name.startswith('module.'):
            name = name[7:]
        index_dict[name] = retained_index.cpu()
        W.data = cuda_pruned_weights  # replace the data field in variable

    np.savez(index_npz, **index_dict)

def admm_initialization(args, ADMM, model):

    if not args.admm:
        return    
    for (name, W) in model.named_parameters():
        if name not in ADMM.prune_cfg:
            continue
        _, updated_Z = weight_pruning(args, W, ADMM.prune_cfg[name])  # Z(k+1) = W(k+1)+U(k)  U(k) is zeros her
        ADMM.ADMM_Z[name] = updated_Z


def z_u_update(args, ADMM, model, train_loader, optimizer, epoch, data, batch_idx):

    if not args.admm:
        return
    if epoch != 1 and (epoch - 1) % args.admm_epochs == 0 and batch_idx == 0:
        for (name, W) in model.named_parameters():
            if name not in ADMM.prune_cfg:
                continue
            W_detach = W.detach()
            U_detach = ADMM.ADMM_U[name].detach()
            ADMM.ADMM_Z[name] = W_detach + U_detach  # Z(k+1) = W(k+1) + U[k]
            _, updated_Z = weight_pruning(args, ADMM.ADMM_Z[name], ADMM.prune_cfg[name])  # equivalent to Euclidean Projection
            ADMM.ADMM_Z[name] = updated_Z
            Z_detach = ADMM.ADMM_Z[name].detach()
            ADMM.ADMM_U[name] = W_detach - Z_detach + U_detach  # U(k+1) = W(k+1) - Z(k+1) + U(k)


def append_admm_loss(args, ADMM, model, ce_loss):
    '''
    append admm loss to cross_entropy loss
    Args:
        args: configuration parameters
        model: instance to the model class
        ce_loss: the cross entropy loss
    Returns:
        ce_loss(tensor scalar): original cross enropy loss
        admm_loss(dict, name->tensor scalar): a dictionary to show loss for each layer
        ret_loss(scalar): the mixed overall loss

    '''
    admm_loss = {}

    if args.admm:
        for i, (name, W) in enumerate(model.named_parameters()):  ## initialize Z (for both weights and bias)
            if name not in ADMM.prune_cfg:
                continue
            # admm_loss[name] = 0.5 * ADMM.rhos[name] * (torch.norm(W - ADMM.ADMM_Z[name] + ADMM.ADMM_U[name], p=2) ** 2)
            admm_loss[name] = 0.5 * ADMM.rhos[name] * (torch.norm((W + ADMM.ADMM_U[name])[ADMM.ADMM_Z[name]==0], p=2) ** 2)
    mixed_loss = 0
    mixed_loss += ce_loss
    for k, v in admm_loss.items():
        mixed_loss += v
    return ce_loss, admm_loss, mixed_loss


def admm_adjust_learning_rate(optimizer, epoch, args):
    ''' (The pytorch learning rate scheduler)
Sets the learning rate to the initial LR decayed by 10 every 30 epochs'''
    '''
    For admm, the learning rate change is periodic.
    When epoch is dividable by admm_epoch, the learning rate is reset
    to the original one, and decay every 3 epoch (as the default 
    admm epoch is 9)

    '''
    admm_epoch = args.admm_epochs
    lr = None
    if (epoch - 1) % admm_epoch == 0:
        lr = args.lr
    else:
        admm_epoch_offset = (epoch - 1) % admm_epoch
        admm_step = admm_epoch / 3  # roughly every 1/3 admm_epoch.
        lr = args.lr * (0.5 ** (admm_epoch_offset // admm_step))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr



# f = open('data/coco/val2017.txt', 'r')
# f1 = open('data/coco/val2017.shapes', 'r')
import os

imagedir = 'data/coco/images/train2017/'
newdir = 'data/coco/empty/train2017'
nonemptyfile = 0
imagelist = os.listdir(newdir)

with open('data/coco/train2017.txt', 'r') as r:
    lines=r.readlines()
with open('data/coco/train2017.shapes', 'r') as r:
    sline=r.readlines()
with open('data/coco/train2020.txt','w') as w:
    with open('data/coco/train2020.shapes', 'w') as ws:
        for i in range(len(lines)):
            imagename = lines[i][-17:-1]
            if imagename not in imagelist:
                w.write(lines[i])
                ws.write(sline[i])
                nonemptyfile = nonemptyfile+1

print(nonemptyfile)



import os

import numpy as np
import lmdb
import six
import torch
from PIL import Image
import torch.utils.data as data
import pyarrow as pa


def random_clip(video_frames, sampling_rate, frames_per_clip, fixed_offset=False):
    """

    Args:
        video_frames (int): total frame number of a video
        sampling_rate (int): sampling rate for clip, pick one every k frames
        frames_per_clip (int): number of frames of a clip
        fixed_offset (bool): used with sample offset to decide the offset value deterministically.

    Returns:
        list[int]: frame indices (started from zero)
    """
    new_sampling_rate = sampling_rate
    highest_idx = video_frames - new_sampling_rate * frames_per_clip
    if highest_idx <= 0:
        random_offset = 0
    else:
        if fixed_offset:
            random_offset = (video_frames - new_sampling_rate * frames_per_clip) // 2
        else:
            random_offset = int(np.random.randint(0, highest_idx, 1))
    frame_idx = [int(random_offset + i * sampling_rate) % video_frames for i in range(frames_per_clip)]
    return frame_idx


class VideoRecord(object):
    def __init__(self, path, start_frame, end_frame, label, reverse=False):
        self.path = path
        self.video_id = os.path.basename(path)
        self.start_frame = start_frame
        self.end_frame = end_frame
        self.label = label
        self.reverse = reverse

    @property
    def num_frames(self):
        return self.end_frame - self.start_frame + 1


class VideoDataSet(data.Dataset):

    def __init__(self, root_path, list_file, num_groups=64, frames_per_group=1, sample_offset=0, num_clips=1,
                 modality='rgb', dense_sampling=False, fixed_offset=True,
                 image_tmpl='{:05d}.jpg', transform=None, is_train=True, test_mode=False, seperator=' ',
                 filter_video=0, num_classes=None, max_frame=None):
        """

        Argments have different meaning when dense_sampling is True:
            - num_groups ==> number of frames
            - frames_per_group ==> sample every K frame
            - sample_offset ==> number of clips used in validation or test mode

        Args:
            root_path (str): the file path to the root of video folder
            list_file (str): the file list, each line with folder_path, start_frame, end_frame, label_id
            num_groups (int): number of frames per data sample
            frames_per_group (int): number of frames within one group
            sample_offset (int): used in validation/test, the offset when sampling frames from a group
            modality (str): rgb or flow
            dense_sampling (bool): dense sampling in I3D
            fixed_offset (bool): used for generating the same videos used in TSM
            image_tmpl (str): template of image ids
            transform: the transformer for preprocessing
            is_train (bool): shuffle the video but keep the causality
            test_mode (bool): testing mode, no label
        """
        if modality not in ['flow', 'rgb']:
            raise ValueError("modality should be 'flow' or 'rgb'.")

        self.root_path = root_path
        self.list_file = list_file
        self.num_groups = num_groups
        self.num_frames = num_groups
        self.frames_per_group = frames_per_group
        self.sample_freq = frames_per_group
        self.num_clips = num_clips
        self.sample_offset = sample_offset
        self.fixed_offset = fixed_offset
        self.dense_sampling = dense_sampling
        self.modality = modality.lower()
        self.image_tmpl = image_tmpl
        self.transform = transform
        self.is_train = is_train
        self.test_mode = test_mode
        self.seperator = seperator
        self.filter_video = filter_video

        self.max_frame = max_frame

        if self.modality == 'flow':
            self.num_consecutive_frames = 5
        else:
            self.num_consecutive_frames = 1

        self.multi_label = None
        self.video_list = self._parse_list()
        self.num_classes = num_classes

    def _load_image(self, directory, idx):

        def _safe_load_image(img_path):
            img_tmp = Image.open(img_path)
            img = img_tmp.copy()
            img_tmp.close()
            return img

        num_try = 0
        image_path_file = os.path.join(self.root_path, directory, self.image_tmpl.format(idx))
        img = None
        while num_try < 10:
            try:
                if self.modality == 'rgb':
                    img = [_safe_load_image(image_path_file)]
                else:
                    ext = image_path_file.split(".")[-1]
                    flow_x_name = image_path_file.replace(".{}".format(ext), "_x.{}".format(ext))
                    flow_y_name = image_path_file.replace(".{}".format(ext), "_y.{}".format(ext))
                    img = [_safe_load_image(flow_x_name), _safe_load_image(flow_y_name)]
                break
            except Exception as e:
                print('[Will try load again] error loading image: {}, error: {}'.format(image_path_file, str(e)))
                num_try += 1

        if img is None:
            raise ValueError('[Fail 10 times] error loading image: {}'.format(image_path_file))

        return img

    def _parse_list(self):
        # usualy it is [video_id, num_frames, class_idx]
        # or [video_id, start_frame, end_frame, list of class_idx]
        tmp = []
        original_video_numbers = 0
        for x in open(self.list_file):
            elements = x.strip().split(self.seperator)
            start_frame = int(elements[1])
            end_frame = int(elements[2])
            total_frame = end_frame - start_frame + 1
            original_video_numbers += 1
            if self.test_mode:
                tmp.append(elements)
            else:
                if total_frame >= self.filter_video:
                    tmp.append(elements)

        num = len(tmp)
        print("The number of videos is {} (with more than {} frames) "
              "(original: {})".format(num, self.filter_video, original_video_numbers), flush=True)
        assert (num > 0)
        # TODO: a better way to check if multi-label or not
        self.multi_label = True if len(tmp[0]) > 4 else False
        file_list = []
        for item in tmp:
            if self.test_mode:
                file_list.append([item[0], int(item[1]), int(item[2]), -1])
            else:
                labels = []
                for i in range(3, len(item)):
                    labels.append(float(item[i]))
                if not self.multi_label:
                    labels = labels[0] if len(labels) == 1 else labels
                file_list.append([item[0], int(item[1]), int(item[2]), labels])

        video_list = [VideoRecord(item[0], item[1], item[2], item[3]) for item in file_list]
        # flow model has one frame less
        if self.modality == 'flow':
            for i in range(len(video_list)):
                video_list[i].end_frame -= 1

        return video_list

    def _sample_indices(self, record):
        """
        Used for training.

        Args:
            - record (VideoRecord):

        Returns:
            list: frame index, index starts from 1.
        """
        max_frame_idx = max(1, record.num_frames - self.num_consecutive_frames + 1)
        freq = max(max_frame_idx // self.max_frame, 1) if self.max_frame is not None else 1
        max_frame_idx = max_frame_idx // freq
        if self.dense_sampling:
            frame_idx = np.asarray(random_clip(max_frame_idx, self.sample_freq, self.num_frames))
        else:
            total_frames = self.num_groups * self.frames_per_group
            ave_frames_per_group = max_frame_idx // self.num_groups
            if ave_frames_per_group >= self.frames_per_group:
                # randomly sample f images per segement
                frame_idx = np.arange(0, self.num_groups) * ave_frames_per_group
                frame_idx = np.repeat(frame_idx, repeats=self.frames_per_group)
                offsets = np.random.choice(ave_frames_per_group, self.frames_per_group, replace=False)
                offsets = np.tile(offsets, self.num_groups)
                frame_idx = frame_idx + offsets
            elif max_frame_idx < total_frames:
                # need to sample the same images
                frame_idx = np.random.choice(max_frame_idx, total_frames)
            else:
                # sample cross all images
                frame_idx = np.random.choice(max_frame_idx, total_frames, replace=False)
            frame_idx = np.sort(frame_idx)
        frame_idx = frame_idx + 1
        return frame_idx

    def _get_val_indices(self, record):
        max_frame_idx = max(1, record.num_frames - self.num_consecutive_frames + 1)
        freq = max(max_frame_idx // self.max_frame, 1) if self.max_frame is not None else 1
        max_frame_idx = max_frame_idx // freq
        if self.dense_sampling:
            if self.fixed_offset:
                sample_pos = max(1, 1 + max_frame_idx - self.sample_freq * self.num_frames)
                t_stride = self.sample_freq
                start_list = np.linspace(0, sample_pos - 1, num=self.num_clips, dtype=int)
                frame_idx = []
                for start_idx in start_list.tolist():
                    frame_idx += [(idx * t_stride + start_idx) % max_frame_idx for idx in range(self.num_frames)]
            else:
                frame_idx = []
                for i in range(self.num_clips):
                    frame_idx.extend(random_clip(max_frame_idx, self.sample_freq, self.num_frames))
            frame_idx = np.asarray(frame_idx) + 1
        else:  # uniform sampling
            if self.fixed_offset:
                frame_idices = []
                sample_offsets = list(range(-self.num_clips // 2 + 1, self.num_clips // 2 + 1))
                for sample_offset in sample_offsets:
                    if max_frame_idx > self.num_groups:
                        tick = max_frame_idx / float(self.num_groups)
                        curr_sample_offset = sample_offset
                        if curr_sample_offset >= tick / 2.0:
                            curr_sample_offset = tick / 2.0 - 1e-4
                        elif curr_sample_offset < -tick / 2.0:
                            curr_sample_offset = -tick / 2.0
                        frame_idx = np.array([int(tick / 2.0 + curr_sample_offset + tick * x) for x in range(self.num_groups)])
                    else:
                        np.random.seed(sample_offset - (-self.num_clips // 2 + 1))
                        frame_idx = np.random.choice(max_frame_idx, self.num_groups)
                    frame_idx = np.sort(frame_idx)
                    frame_idices.extend(frame_idx.tolist())
            else:
                frame_idices = []
                for i in range(self.num_clips):
                    total_frames = self.num_groups * self.frames_per_group
                    ave_frames_per_group = max_frame_idx // self.num_groups
                    if ave_frames_per_group >= self.frames_per_group:
                        # randomly sample f images per segment
                        frame_idx = np.arange(0, self.num_groups) * ave_frames_per_group
                        frame_idx = np.repeat(frame_idx, repeats=self.frames_per_group)
                        offsets = np.random.choice(ave_frames_per_group, self.frames_per_group, replace=False)
                        offsets = np.tile(offsets, self.num_groups)
                        frame_idx = frame_idx + offsets
                    elif max_frame_idx < total_frames:
                        # need to sample the same images
                        np.random.seed(i)
                        frame_idx = np.random.choice(max_frame_idx, total_frames)
                    else:
                        # sample cross all images
                        np.random.seed(i)
                        frame_idx = np.random.choice(max_frame_idx, total_frames, replace=False)
                    frame_idx = np.sort(frame_idx)
                    frame_idices.extend(frame_idx.tolist())
            frame_idx = np.asarray(frame_idices) + 1
        return frame_idx

    def __getitem__(self, index):
        """
        Returns:
            torch.FloatTensor: (3xgxf)xHxW dimension, g is number of groups and f is the frames per group.
            torch.FloatTensor: the label
        """
        record = self.video_list[index]
        # check this is a legit video folder
        if self.is_train:
            indices = self._sample_indices(record)
        else:
            indices = self._get_val_indices(record)

        images = []
        for seg_ind in indices:
            for i in range(self.num_consecutive_frames):
                new_seg_ind = min(seg_ind + record.start_frame - 1 + i, record.num_frames)
                seg_imgs = self._load_image(record.path, new_seg_ind)
                images.extend(seg_imgs)

        images = self.transform(images)
        if self.test_mode:
            # in test mode, return the video id as label
            label = int(record.video_id)
        else:
            if not self.multi_label:
                label = int(record.label)
            else:
                # create a binary vector.
                label = torch.zeros(self.num_classes, dtype=torch.float)
                for x in record.label:
                    label[int(x)] = 1.0

        # re-order data to targeted format.
        return images, label

    def __len__(self):
        return len(self.video_list)


class VideoDataSetLMDB(data.Dataset):

    def __init__(self, root_path, list_file, num_groups=64, frames_per_group=1, sample_offset=0, num_clips=1,
                 modality='rgb', dense_sampling=False, fixed_offset=True,
                 image_tmpl='{:05d}.jpg', transform=None, is_train=True, test_mode=False,
                 seperator=' ', filter_video=0, num_classes=None, max_frame=None):
        """

        Arguments have different meaning when dense_sampling is True:
            - num_groups ==> number of frames
            - frames_per_group ==> sample every K frame
            - sample_offset ==> number of clips used in validation or test mode

        Args:
            root_path (str): the file path to the root of video folder
            num_groups (int): number of frames per data sample
            frames_per_group (int): number of frames within one group
            sample_offset (int): used in validation/test, the offset when sampling frames from a group
            modality (str): rgb or flow
            dense_sampling (bool): dense sampling in I3D
            fixed_offset (bool): used for generating the same videos used in TSM
            image_tmpl (str): template of image ids
            transform: the transformer for preprocessing
            is_train (bool): shuffle the video but keep the causality
            test_mode (bool): testing mode, no label
        """
        # TODO: handle multi-label?
        # TODO: flow data?

        if modality not in ['flow', 'rgb']:
            raise ValueError("modality should be 'flow' or 'rgb'.")

        self.root_path = root_path
        self.num_groups = num_groups
        self.num_frames = num_groups
        self.frames_per_group = frames_per_group
        self.sample_freq = frames_per_group
        self.num_clips = num_clips
        self.sample_offset = sample_offset
        self.fixed_offset = fixed_offset
        self.dense_sampling = dense_sampling
        self.modality = modality.lower()
        self.image_tmpl = image_tmpl
        self.transform = transform
        self.is_train = is_train
        self.test_mode = test_mode
        self.seperator = seperator
        self.filter_video = filter_video

        self.max_frame = max_frame

        if self.modality == 'flow':
            self.num_consecutive_frames = 5
        else:
            self.num_consecutive_frames = 1

        self.multi_label = None

        self.db = None
        db = lmdb.open(self.root_path, max_readers=1, subdir=os.path.isdir(self.root_path),
                       readonly=True, lock=False, readahead=False, meminit=False)
        with db.begin(write=False) as txn:
            self.length = pa.deserialize(txn.get(b'__len__'))
            self.keys = pa.deserialize(txn.get(b'__keys__'))
        db.close()

        # TODO: a hack way to filter video
        # self.list_file = root_path.replace(".lmdb", ".txt")
        self.list_file = list_file

        valid_video_numbers = 0
        invalid_video_ids = []
        for x in open(self.list_file):
            elements = x.strip().split(self.seperator)
            start_frame = int(elements[1])
            end_frame = int(elements[2])
            total_frame = end_frame - start_frame + 1
            if self.test_mode:
                valid_video_numbers += 1
            else:
                if total_frame >= self.filter_video:
                    valid_video_numbers += 1
                else:
                    name = u'{}'.format(elements[0].split("/")[-1]).encode('ascii')
                    invalid_video_ids.append(name)

        print("The number of videos is {} (with more than {} frames) "
              "(original: {})".format(valid_video_numbers, self.filter_video, self.length),
              flush=True)

        # remove keys and update length
        self.length = valid_video_numbers
        self.keys = [k for k in self.keys if k not in invalid_video_ids]

        if self.length != len(self.keys):
            raise ValueError("Do not filter video correctly.")

        self.num_classes = num_classes

    def _sample_indices(self, record):
        """
        Used for training.

        Args:
            - record (VideoRecord):

        Returns:
            list: frame index, index starts from 1.
        """
        max_frame_idx = max(1, record.num_frames - self.num_consecutive_frames + 1)
        freq = max(max_frame_idx // self.max_frame, 1) if self.max_frame is not None else 1
        max_frame_idx = max_frame_idx // freq
        if self.dense_sampling:
            frame_idx = np.asarray(random_clip(max_frame_idx, self.sample_freq, self.num_frames))
        else:
            total_frames = self.num_groups * self.frames_per_group
            ave_frames_per_group = max_frame_idx // self.num_groups
            if ave_frames_per_group >= self.frames_per_group:
                # randomly sample f images per segement
                frame_idx = np.arange(0, self.num_groups) * ave_frames_per_group
                frame_idx = np.repeat(frame_idx, repeats=self.frames_per_group)
                offsets = np.random.choice(ave_frames_per_group, self.frames_per_group, replace=False)
                offsets = np.tile(offsets, self.num_groups)
                frame_idx = frame_idx + offsets
            elif max_frame_idx < total_frames:
                # need to sample the same images
                frame_idx = np.random.choice(max_frame_idx, total_frames)
            else:
                # sample cross all images
                frame_idx = np.random.choice(max_frame_idx, total_frames, replace=False)
            frame_idx = np.sort(frame_idx)
        frame_idx = frame_idx + 1
        return frame_idx

    def _get_val_indices(self, record):
        max_frame_idx = max(1, record.num_frames - self.num_consecutive_frames + 1)
        freq = max(max_frame_idx // self.max_frame, 1) if self.max_frame is not None else 1
        max_frame_idx = max_frame_idx // freq
        if self.dense_sampling:
            if self.fixed_offset:
                sample_pos = max(1, 1 + max_frame_idx - self.sample_freq * self.num_frames)
                t_stride = self.sample_freq
                start_list = np.linspace(0, sample_pos - 1, num=self.num_clips, dtype=int)
                frame_idx = []
                for start_idx in start_list.tolist():
                    frame_idx += [(idx * t_stride + start_idx) % max_frame_idx for idx in range(self.num_frames)]
            else:
                frame_idx = []
                for i in range(self.num_clips):
                    frame_idx.extend(random_clip(max_frame_idx, self.sample_freq, self.num_frames))
            frame_idx = np.asarray(frame_idx) + 1
        else:  # uniform sampling
            if self.fixed_offset:
                frame_idices = []
                sample_offsets = list(range(-self.num_clips // 2 + 1, self.num_clips // 2 + 1))
                for sample_offset in sample_offsets:
                    if max_frame_idx > self.num_groups:
                        tick = max_frame_idx / float(self.num_groups)
                        curr_sample_offset = sample_offset
                        if curr_sample_offset >= tick / 2.0:
                            curr_sample_offset = tick / 2.0 - 1e-4
                        elif curr_sample_offset < -tick / 2.0:
                            curr_sample_offset = -tick / 2.0
                        frame_idx = np.array([int(tick / 2.0 + curr_sample_offset + tick * x) for x in range(self.num_groups)])
                    else:
                        np.random.seed(sample_offset - (-self.num_clips // 2 + 1))
                        frame_idx = np.random.choice(max_frame_idx, self.num_groups)
                    frame_idx = np.sort(frame_idx)
                    frame_idices.extend(frame_idx.tolist())
            else:
                frame_idices = []
                for i in range(self.num_clips):
                    total_frames = self.num_groups * self.frames_per_group
                    ave_frames_per_group = max_frame_idx // self.num_groups
                    if ave_frames_per_group >= self.frames_per_group:
                        # randomly sample f images per segment
                        frame_idx = np.arange(0, self.num_groups) * ave_frames_per_group
                        frame_idx = np.repeat(frame_idx, repeats=self.frames_per_group)
                        offsets = np.random.choice(ave_frames_per_group, self.frames_per_group, replace=False)
                        offsets = np.tile(offsets, self.num_groups)
                        frame_idx = frame_idx + offsets
                    elif max_frame_idx < total_frames:
                        # need to sample the same images
                        np.random.seed(i)
                        frame_idx = np.random.choice(max_frame_idx, total_frames)
                    else:
                        # sample cross all images
                        np.random.seed(i)
                        frame_idx = np.random.choice(max_frame_idx, total_frames, replace=False)
                    frame_idx = np.sort(frame_idx)
                    frame_idices.extend(frame_idx.tolist())
            frame_idx = np.asarray(frame_idices) + 1
        return frame_idx

    def __getitem__(self, index):
        if self.db is None:
            self.db = lmdb.open(self.root_path, max_readers=1, subdir=os.path.isdir(self.root_path),
                                readonly=True, lock=False, readahead=False, meminit=False)

        with self.db.begin(write=False) as txn:
            byteflow = txn.get(self.keys[index])
        try:
            unpacked_video = pa.deserialize(byteflow)
        except:
            print('error in loading {}, {}'.format(index, self.keys[index]))
            return None

        record = VideoRecord("", 1, unpacked_video[0], unpacked_video[-1])
        if self.is_train:
            indices = self._sample_indices(record)
        else:
            indices = self._get_val_indices(record)

        # load image
        def _safe_load_image(buf):
            img_tmp = Image.open(buf).convert('RGB')
            img = img_tmp.copy()
            img_tmp.close()
            return img

        images = []
        for seg_ind in indices:
            for i in range(self.num_consecutive_frames):
                new_seg_ind = min(seg_ind + record.start_frame - 1 + i, record.num_frames)
                imgbuf = unpacked_video[new_seg_ind]
                buf = six.BytesIO()
                buf.write(imgbuf)
                buf.seek(0)
                img = _safe_load_image(buf)
                images.extend([img])
        # import cv2; cv2.imwrite('miao.jpg', np.array(images[0]))
        images = self.transform(images)

        if self.test_mode:
            # in test mode, return the video id as label
            label = int(self.keys[index])
        else:
            if not self.multi_label:
                label = int(record.label)
            else:
                # create a binary vector.
                label = torch.zeros(self.num_classes, dtype=torch.float)
                for x in record.label:
                    label[int(x)] = 1.0
        
        # re-order data to targeted format.
        return images, label
        
    def __len__(self):
        return self.length

    


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo

from network.inflate_from_2d_model import inflate_from_2d_model

__all__ = ['s3d']

model_urls = {
    'googlenet': 'https://download.pytorch.org/models/googlenet-1378be20.pth',
}


'''
A pytorch implementation of S3D based on the code in https://github.com/tensorflow/models/blob/master/research/slim/nets/s3dg.py.
One difference in this implementation (more costly) is that it performs temporal pooling in the last three layers while S3DG performs does so
in the first convolution and the last two layers.
'''
class S3D(nn.Module):
    """
    reference: https://github.com/tensorflow/models/blob/master/research/slim/nets/s3dg.py
    """
    def __init__(self, num_classes=1000, dropout_ratio=0.2, without_t_stride=False,
                 pooling_method='max', dw_t_conv=False):
        super(S3D, self).__init__()
        self.pooling_method = pooling_method.lower()
        if self.pooling_method == 'avg':
            self.pooling_functor = F.avg_pool3d
        else:
            self.pooling_functor = F.max_pool3d
        self.dw_t_conv = dw_t_conv
        self.without_t_stride = without_t_stride
        self.t_s = 1 if without_t_stride else 2
        # no temporal downsampling at the first layer
        self.conv1 = BasicConv3d(3, 64, kernel_size=(3, 7, 7), stride=(1, 2, 2), padding=(1, 3, 3)) # (64, 32, 112, 112)
        self.conv2 = BasicConv3d(64, 64, kernel_size=1)
        self.conv3 = STConv3d(64, 192, kernel_size=3, stride=1, padding=1,
                              dw_t_conv=dw_t_conv)

        self.inception3a = Inception(192, 64, 96, 128, 16, 32, 32, dw_t_conv)
        self.inception3b = Inception(256, 128, 128, 192, 32, 96, 64, dw_t_conv)

        self.inception4a = Inception(480, 192, 96, 208, 16, 48, 64, dw_t_conv)
        self.inception4b = Inception(512, 160, 112, 224, 24, 64, 64, dw_t_conv)
        self.inception4c = Inception(512, 128, 128, 256, 24, 64, 64, dw_t_conv)
        self.inception4d = Inception(512, 112, 144, 288, 32, 64, 64, dw_t_conv)
        self.inception4e = Inception(528, 256, 160, 320, 32, 128, 128, dw_t_conv)

        self.inception5a = Inception(832, 256, 160, 320, 32, 128, 128, dw_t_conv)
        self.inception5b = Inception(832, 384, 192, 384, 48, 128, 128, dw_t_conv)

        self.dropout = nn.Dropout(dropout_ratio)
    #    self.fc = nn.Conv3d(1024, num_classes, kernel_size=1, stride=1, bias=True)
        self.fc = nn.Linear(1024, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm3d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.001)
                nn.init.constant_(m.bias, 0)

        # initial temporal conv with 1 at center frame and other frames 0
        # for m in self.modules():
        #     if isinstance(m, STConv3d):
        #         tmp_weight = torch.zeros(m.conv_t.weight.shape)
        #         center_pos = tmp_weight.shape[2] // 2
        #         tmp_weight[:, :, center_pos, 0, 0] = 1.0
        #         m.conv_t.weight = nn.Parameter(tmp_weight)

    def mean(self, modality='rgb'):
        return [0.5, 0.5, 0.5] if modality == 'rgb' else [0.5]

    def std(self, modality='rgb'):
        return [0.5, 0.5, 0.5] if modality == 'rgb' else [0.5]

    @property
    def network_name(self):
        name = 's3d'
        if self.dw_t_conv:
            name += '-dw-t-conv'
        if not self.without_t_stride:
            name += '-ts-{}'.format(self.pooling_method)

        return name

    def forward(self, x):
        # N x 3 x F x 224 x 224
        x = self.conv1(x)
        # N x 64 x F x 112 x 112
        x = self.pooling_functor(x, kernel_size=(1, 3, 3), stride=(1, 2, 2), padding=(0, 1, 1))
        # N x 64 x F x 56 x 56
        x = self.conv2(x)
        # N x 64 x F x 56 x 56
        x = self.conv3(x)
        # N x 192 x F x 56 x 56
        # difference from the original S3D , which does temporal downsampling in the first conv
        x = self.pooling_functor(x, kernel_size=(3, 3, 3), stride=(self.t_s, 2, 2), padding=(1, 1, 1))
        # N x 192 x (F/2) x 28 x 28
        x = self.inception3a(x)
        # N x 256 x (F/2) x 28 x 28
        x = self.inception3b(x)
        # N x 480 x (F/2) x 28 x 28
        x = self.pooling_functor(x, kernel_size=3, stride=(self.t_s, 2, 2), padding=1)
        # N x 480 x (F/4) x 14 x 14
        x = self.inception4a(x)
        # N x 512 x (F/4) x 14 x 14
        x = self.inception4b(x)
        # N x 512 x (F/4) x 14 x 14
        x = self.inception4c(x)
        # N x 512 x (F/4) x 14 x 14
        x = self.inception4d(x)
        # N x 528 x (F/4) x 14 x 14
        x = self.inception4e(x)
        # N x 832 x (F/4) x 14 x 14
        x = self.pooling_functor(x, kernel_size=2, stride=(self.t_s, 2, 2))
        # N x 832 x (F/8) x 7 x 7
        x = self.inception5a(x)
        # N x 832 x (F/8) x 7 x 7
        x = self.inception5b(x)
        # N x 1024 x (F/8) x 7 x 7
        num_frames = x.shape[2]
        x = F.adaptive_avg_pool3d(x, output_size=(num_frames, 1, 1))
        # N x 1024 x ((F/8)-1) x 1 x 1
        x = x.squeeze(-1)
        x = x.squeeze(-1)
        x = x.transpose(1, 2)
        n, c, nf = x.size()
        x = x.contiguous().view(n * c, -1)
        x = self.dropout(x)
        x = self.fc(x)
        x = x.view(n, c, -1)
        # N x num_classes x ((F/8)-1)
        logits = torch.mean(x, 1)
        # squeeze spatial dim
        '''
        x = self.dropout(x)
        x = self.fc(x)
        # squeeze spatial dim
        x = x.squeeze(-1)
        x = x.squeeze(-1)
        # N x num_classes x ((F/8)-1)
        logits = torch.mean(x, 2)
        '''
        # N x num_classes
        return logits


class Inception(nn.Module):

    def __init__(self, in_channels, ch1x1, ch3x3red, ch3x3, ch5x5red, ch5x5, pool_proj,
                 dw_t_conv):
        super(Inception, self).__init__()

        self.branch1 = BasicConv3d(in_channels, ch1x1, kernel_size=1)

        self.branch2 = nn.Sequential(
            BasicConv3d(in_channels, ch3x3red, kernel_size=1),
            STConv3d(ch3x3red, ch3x3, kernel_size=3, stride=1, padding=1, dw_t_conv=dw_t_conv)
        )

        self.branch3 = nn.Sequential(
            BasicConv3d(in_channels, ch5x5red, kernel_size=1),
            STConv3d(ch5x5red, ch5x5, kernel_size=3, stride=1, padding=1, dw_t_conv=dw_t_conv)
        )

        self.branch4 = nn.Sequential(
            nn.MaxPool3d(kernel_size=(3, 3, 3), stride=1, padding=1, ceil_mode=True),
            BasicConv3d(in_channels, pool_proj, kernel_size=1)
        )

    def forward(self, x):
        branch1 = self.branch1(x)
        branch2 = self.branch2(x)
        branch3 = self.branch3(x)
        branch4 = self.branch4(x)

        outputs = [branch1, branch2, branch3, branch4]
        return torch.cat(outputs, 1)


class STConv3d(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride, padding=0, dw_t_conv=False):
        super(STConv3d, self).__init__()
        self.conv = nn.Conv3d(in_planes, out_planes, kernel_size=(1, kernel_size, kernel_size),
                              stride=(1, stride, stride), padding=(0, padding, padding), bias=False)
        self.bn = nn.BatchNorm3d(out_planes, eps=1e-3)
        self.relu = nn.ReLU(inplace=True)

        self.conv_t = nn.Conv3d(out_planes, out_planes, kernel_size=(kernel_size, 1, 1),
                                stride=(stride, 1, 1), padding=(padding, 0, 0),
                                groups=out_planes if dw_t_conv else 1)
        self.bn_t = nn.BatchNorm3d(out_planes, eps=1e-3)
        self.relu_t = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        x = self.conv_t(x)
        x = self.bn_t(x)
        x = self.relu_t(x)
        return x


class BasicConv3d(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride=(1, 1, 1), padding=(0, 0, 0)):
        super(BasicConv3d, self).__init__()
        self.conv = nn.Conv3d(in_planes, out_planes, kernel_size=kernel_size,
                              stride=stride, padding=padding, bias=False)

        self.bn = nn.BatchNorm3d(out_planes, eps=1e-3)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


def s3d(num_classes, dropout, without_t_stride, pooling_method, dw_t_conv, **kwargs):
    model = S3D(num_classes=num_classes, dropout_ratio=dropout, without_t_stride=without_t_stride,
                pooling_method=pooling_method, dw_t_conv=dw_t_conv)
    new_model_state_dict = model.state_dict()
    state_dict = model_zoo.load_url(model_urls['googlenet'], map_location='cpu', progress=True)
    state_d = inflate_from_2d_model(state_dict, new_model_state_dict,
                                    skipped_keys=['fc', 'aux1', 'aux2'])
    model.load_state_dict(state_d, strict=False)
    return model


if __name__ == '__main__':
    from torchsummary import torchsummary
    model = s3d(400, 0.5, without_t_stride=True, pooling_method='max', dw_t_conv=False)

    dummy_data = (3, 16, 224, 224)
    model.eval()
    model_summary = torchsummary.summary(model, input_size=dummy_data)
    print(model.network_name)



# This file contains google utils: https://cloud.google.com/storage/docs/reference/libraries
# pip install --upgrade google-cloud-storage

import os
import time


# from google.cloud import storage


def gdrive_download(id='1HaXkef9z6y5l4vUnCYgdmEAj61c6bfWO', name='coco.zip'):
    # https://gist.github.com/tanaikech/f0f2d122e05bf5f971611258c22c110f
    # Downloads a file from Google Drive, accepting presented query
    # from utils.google_utils import *; gdrive_download()
    t = time.time()

    print('Downloading https://drive.google.com/uc?export=download&id=%s as %s... ' % (id, name), end='')
    os.remove(name) if os.path.exists(name) else None  # remove existing
    os.remove('cookie') if os.path.exists('cookie') else None

    # Attempt file download
    os.system("curl -c ./cookie -s -L \"https://drive.google.com/uc?export=download&id=%s\" > /dev/null" % id)
    if os.path.exists('cookie'):  # large file
        s = "curl -Lb ./cookie \"https://drive.google.com/uc?export=download&confirm=`awk '/download/ {print $NF}' ./cookie`&id=%s\" -o %s" % (
            id, name)
    else:  # small file
        s = "curl -s -L -o %s 'https://drive.google.com/uc?export=download&id=%s'" % (name, id)
    r = os.system(s)  # execute, capture return values
    os.remove('cookie') if os.path.exists('cookie') else None

    # Error check
    if r != 0:
        os.remove(name) if os.path.exists(name) else None  # remove partial
        print('Download error ')  # raise Exception('Download error')
        return r

    # Unzip if archive
    if name.endswith('.zip'):
        print('unzipping... ', end='')
        os.system('unzip -q %s' % name)  # unzip
        os.remove(name)  # remove zip to free space

    print('Done (%.1fs)' % (time.time() - t))
    return r


def upload_blob(bucket_name, source_file_name, destination_blob_name):
    # Uploads a file to a bucket
    # https://cloud.google.com/storage/docs/uploading-objects#storage-upload-object-python

    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)

    blob.upload_from_filename(source_file_name)

    print('File {} uploaded to {}.'.format(
        source_file_name,
        destination_blob_name))


def download_blob(bucket_name, source_blob_name, destination_file_name):
    # Uploads a blob from a bucket
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(source_blob_name)

    blob.download_to_filename(destination_file_name)

    print('Blob {} downloaded to {}.'.format(
        source_blob_name,
        destination_file_name))
